from .autocare_dlt import export_autocare_dlt
from .coco import export_coco
from .transformers import export_transformers
from .yolo import export_yolo

__all__ = ["export_yolo", "export_coco", "export_transformers", "export_autocare_dlt"]


import inspect


class BaseCLI:
    def __init__(self):

        self.methods = []
        self.register_methods()

    def get_object(self):
        raise NotImplementedError("get_object method is not implemented.")

    def register_methods(self):
        class_object = self.get_object()

        self.methods = [
            function_name
            for function_name, _ in inspect.getmembers(class_object)
            if not function_name.startswith("_")
        ]
        for method in self.methods:
            setattr(self, method, getattr(class_object, method))


def cli(_class, _instance):
    def switch_type(command: str = None, **kwargs) -> str:
        name = kwargs.get("name", None)

        class_method_names = [
            function_name
            for function_name, _ in inspect.getmembers(_class, predicate=inspect.ismethod)
        ]
        instance_method_names = [
            function_name
            for function_name, _ in inspect.getmembers(_class, predicate=inspect.isfunction)
            if function_name not in class_method_names and not function_name.startswith("_")
        ]
        attribute_names = [
            attribute_name
            for attribute_name, _ in inspect.getmembers(_class)
            if not attribute_name.startswith("_")
            and attribute_name not in (class_method_names + instance_method_names)
        ]

        help_string = (
            "Available Class Methodas:\n"
            + "\n".join(class_method_names)
            + "\n\nAvailable Instance Methods (need to specify --name [Name]):\n"
            + "\n".join(instance_method_names)
            + "\n\nAvailable Attributes: (need to specify --name [Name])\n"
            + "\n".join(attribute_names)
        )

        if command is None or command == "help":
            return help_string

        if kwargs.get("help", False):
            if command in class_method_names:
                return getattr(_class, command).__doc__
            elif command in instance_method_names + attribute_names:
                return (
                    "You need to specify --name [Name] --root_dir [Root Directory of the instance (optional)]\n"
                    + "to use instance_method or to get attribute\n"
                    + getattr(_class, command).__doc__
                )

        elif command in class_method_names:
            try:
                return getattr(_class, command)(**kwargs)
            except Exception as e:
                raise e

        elif command in instance_method_names:
            name = kwargs.pop("name", None)
            if name is None:
                raise ValueError("You need to specify --name [Name]")

            root_dir = kwargs.pop("root_dir", None)
            instance = _instance(name, root_dir=root_dir)

            try:
                return getattr(instance, command)(**kwargs)
            except Exception as e:
                raise e

        elif command in attribute_names:
            name = kwargs.pop("name", None)
            if name is None:
                raise ValueError("You need to specify --name [Name]")

            root_dir = kwargs.pop("root_dir", None)
            instance = _instance(name, root_dir=root_dir)

            try:
                return getattr(instance, command)
            except Exception as e:
                raise e
        else:
            raise ValueError(f"Command {command} does not exist.")

    return switch_type



from .configs import ModelConfig, TrainConfig, InferenceConfig, ExportConfig
from .data import DatasetInfo, ImageInfo
from .fields import Image, Category, Annotation

__all__ = [
    "ModelConfig",
    "TrainConfig",
    "InferenceConfig",
    "ExportConfig",
    "DatasetInfo",
    "ImageInfo",
    "Image",
    "Category",
    "Annotation",
]



from waffle_hub import TaskType
from waffle_hub.schema.configs import TrainConfig

# Common
MODEL_TYPES = {
    TaskType.OBJECT_DETECTION: {"yolov8": list("nsmlx")},
    TaskType.CLASSIFICATION: {"yolov8": list("nsmlx")},
    TaskType.INSTANCE_SEGMENTATION: {"yolov8": list("nsmlx")},
    # "keypoint_detection": {"yolov8": list("nsmlx")},
}

# Backend Specifics
TASK_MAP = {
    TaskType.OBJECT_DETECTION: "detect",
    TaskType.CLASSIFICATION: "classify",
    TaskType.INSTANCE_SEGMENTATION: "segment"
    # "keypoint_detection": "pose"
}
TASK_SUFFIX = {
    "detect": "",
    "classify": "-cls",
    "segment": "-seg",
}

DEFAULT_PARAMS = {
    TaskType.OBJECT_DETECTION: {
        "yolov8": {
            "n": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=64,
            ),
            "s": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=32,
            ),
            "m": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=16,
            ),
            "l": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=8,
            ),
            "x": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=8,
            ),
        }
    },
    TaskType.CLASSIFICATION: {
        "yolov8": {
            "n": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=512,
            ),
            "s": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=256,
            ),
            "m": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=128,
            ),
            "l": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=64,
            ),
            "x": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=0.01,
                letter_box=False,
                batch_size=64,
            ),
        }
    },
    TaskType.INSTANCE_SEGMENTATION: {
        "yolov8": {
            "n": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=32,
            ),
            "s": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=16,
            ),
            "m": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=8,
            ),
            "l": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=4,
            ),
            "x": TrainConfig(
                epochs=50,
                image_size=[640, 640],
                learning_rate=0.01,
                letter_box=True,
                batch_size=4,
            ),
        }
    },
}



"""
Transformers Hub
See Hub documentation for more details about usage.
"""

import os
import warnings
from copy import deepcopy
from functools import cached_property
from pathlib import Path
from typing import Callable, Union

import torch
import transformers
from torchvision import transforms as T
from transformers import (
    AutoImageProcessor,
    AutoModelForImageClassification,
    AutoModelForObjectDetection,
    Trainer,
    TrainerCallback,
)
from transformers.utils import ModelOutput
from waffle_utils.file import io

from datasets import load_from_disk
from waffle_hub import TaskType
from waffle_hub.hub import Hub
from waffle_hub.hub.adapter.transformers.train_input_helper import (
    ClassifierInputHelper,
    ObjectDetectionInputHelper,
    customTrainingArguments,
)
from waffle_hub.hub.model.wrapper import ModelWrapper
from waffle_hub.schema.configs import TrainConfig
from waffle_hub.utils.callback import TrainCallback

from .config import DEFAULT_PARAMS, MODEL_TYPES


class CustomCallback(TrainerCallback):
    """
    This class is necessary to obtain logs for the training.
    """

    def __init__(self, trainer) -> None:
        super().__init__()
        self._trainer = trainer

    def on_epoch_end(self, args, state, control, **kwargs):
        if control.should_evaluate:
            control_copy = deepcopy(control)
            self._trainer.evaluate(
                eval_dataset=self._trainer.train_dataset,
                metric_key_prefix="train",
            )
            return control_copy


class TransformersHub(Hub):
    BACKEND_NAME = "transformers"
    MODEL_TYPES = MODEL_TYPES
    MULTI_GPU_TRAIN = False
    DEFAULT_PARAMS = DEFAULT_PARAMS

    # Override
    LAST_CKPT_FILE = "weights/last_ckpt"
    BEST_CKPT_FILE = "weights/best_ckpt"

    def __init__(
        self,
        name: str,
        task: str,
        model_type: str,
        model_size: str,
        categories: Union[list[dict], list] = None,
        root_dir: str = None,
        backend: str = None,
        version: str = None,
        *args,
        **kwargs,
    ):
        if backend is not None and TransformersHub.BACKEND_NAME != backend:
            raise ValueError(
                f"Backend {backend} is not supported. Please use {TransformersHub.BACKEND_NAME}"
            )

        if version is not None and transformers.__version__ != version:
            warnings.warn(
                f"You've loaded the Hub created with transformers=={version}, \n"
                + f"but the installed version is {transformers.__version__}."
            )

        super().__init__(
            name=name,
            backend=TransformersHub.BACKEND_NAME,
            version=transformers.__version__,
            task=task,
            model_type=model_type,
            model_size=model_size,
            categories=categories,
            root_dir=root_dir,
        )

    # Override
    @cached_property
    def best_ckpt_file(self) -> Path:
        """Best Checkpoint File"""
        return self.hub_dir / TransformersHub.BEST_CKPT_FILE

    @cached_property
    def last_ckpt_file(self) -> Path:
        """Last Checkpoint File"""
        return self.hub_dir / TransformersHub.LAST_CKPT_FILE

    @classmethod
    def new(
        cls,
        name: str,
        task: str = None,
        model_type: str = None,
        model_size: str = None,
        categories: Union[list[dict], list] = None,
        root_dir: str = None,
        *args,
        **kwargs,
    ):

        warnings.warn("UltralyticsHub.new() is deprecated. Please use Hub.new() instead.")

        return cls(
            name=name,
            task=task,
            model_type=model_type,
            model_size=model_size,
            categories=categories,
            root_dir=root_dir,
        )

    def on_train_start(self, cfg: TrainConfig):
        # overwrite train config with default config
        cfg.pretrained_model = self.MODEL_TYPES[self.task][self.model_type][self.model_size]

        dataset = load_from_disk(cfg.dataset_path)

        if self.task == TaskType.CLASSIFICATION:
            helper = ClassifierInputHelper(cfg.pretrained_model, cfg.image_size)
            cfg.train_input = helper.get_train_input()
            categories = dataset["train"].features["label"].names
            id2label = {index: x for index, x in enumerate(categories, start=0)}
            cfg.train_input.model = AutoModelForImageClassification.from_pretrained(
                cfg.pretrained_model,
                num_labels=len(id2label),
                ignore_mismatched_sizes=True,
            )

        elif self.task == TaskType.OBJECT_DETECTION:
            helper = ObjectDetectionInputHelper(cfg.pretrained_model, cfg.image_size)
            cfg.train_input = helper.get_train_input()
            categories = dataset["train"].features["objects"].feature["category"].names
            id2label = {index: x for index, x in enumerate(categories, start=0)}
            label2id = {x: index for index, x in enumerate(categories, start=0)}
            cfg.train_input.model = AutoModelForObjectDetection.from_pretrained(
                cfg.pretrained_model,
                id2label=id2label,
                label2id=label2id,
                ignore_mismatched_sizes=True,
            )
        else:
            raise NotImplementedError

        if cfg.device == "cpu":
            os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
            cfg.train_input.model = cfg.train_input.model.to("cpu")
        elif "," in cfg.device:
            os.environ["CUDA_VISIBLE_DEVICES"] = cfg.device

        transforms = helper.get_transforms()
        dataset["train"] = dataset["train"].with_transform(transforms)
        dataset["val"] = dataset["val"].with_transform(transforms)
        cfg.train_input.dataset = dataset

        cfg.train_input.training_args = customTrainingArguments(
            output_dir=str(self.artifact_dir),
            per_device_train_batch_size=cfg.batch_size,
            num_train_epochs=cfg.epochs,
            remove_unused_columns=False,
            push_to_hub=False,
            logging_strategy="epoch" if cfg.verbose else "no",
            evaluation_strategy="epoch",
            save_strategy="epoch",
            learning_rate=cfg.learning_rate,
            dataloader_num_workers=cfg.workers,
            seed=cfg.seed,
            metric_for_best_model="eval_loss",
            greater_is_better=False,
            save_total_limit=1,
            load_best_model_at_end=False,
            device=cfg.device,
        )

    def training(self, cfg: TrainConfig, callback: TrainCallback):
        trainer = Trainer(
            model=cfg.train_input.model,
            args=cfg.train_input.training_args,
            data_collator=cfg.train_input.collator,
            train_dataset=cfg.train_input.dataset["train"],
            eval_dataset=cfg.train_input.dataset["val"],
            tokenizer=cfg.train_input.image_processor,
            compute_metrics=cfg.train_input.compute_metrics,
        )
        trainer.add_callback(CustomCallback(trainer))
        trainer.train()
        trainer.save_model(str(self.artifact_dir / "weights" / "last_ckpt"))
        trainer._load_best_model()
        trainer.save_model(str(self.artifact_dir / "weights" / "best_ckpt"))

        self.train_log = trainer.state.log_history

    def get_metrics(self) -> list[list[dict]]:
        metrics = []

        for epoch in range(0, len(self.train_log) - 1, 3):  # last is runtime info

            current_epoch_log = self.train_log[epoch]

            current_epoch_log.update(self.train_log[epoch + 1])
            current_epoch_log.update(self.train_log[epoch + 2])

            epoch_metrics = []
            for key, value in current_epoch_log.items():
                epoch_metrics.append({"tag": key, "value": value})
            metrics.append(epoch_metrics)

        return metrics

    def on_train_end(self, cfg: TrainConfig):
        io.copy_files_to_directory(
            self.artifact_dir / "weights" / "best_ckpt",
            self.best_ckpt_file,
            create_directory=True,
        )
        io.copy_files_to_directory(
            self.artifact_dir / "weights" / "last_ckpt",
            self.last_ckpt_file,
            create_directory=True,
        )
        io.save_json(self.get_metrics(), self.metric_file)

    def get_preprocess(self, pretrained_model: str = None) -> Callable:
        if pretrained_model is None:
            pretrained_model = self.best_ckpt_file
        image_processer = AutoImageProcessor.from_pretrained(pretrained_model)

        normalize = T.Normalize(image_processer.image_mean, image_processer.image_std, inplace=True)

        def preprocess(x, *args, **kwargs):
            return normalize(x)

        return preprocess

    def get_postprocess(self: str, pretrained_model: str = None) -> Callable:
        if pretrained_model is None:
            pretrained_model = self.best_ckpt_file
        image_processer = AutoImageProcessor.from_pretrained(pretrained_model)

        if self.task == TaskType.CLASSIFICATION:

            def inner(x: ModelOutput, *args, **kwargs) -> torch.Tensor:
                return [x.logits]

        elif self.task == TaskType.OBJECT_DETECTION:
            post_process = image_processer.post_process_object_detection

            def inner(x: ModelOutput, *args, **kwargs) -> torch.Tensor:

                x = post_process(x, threshold=-1)

                xyxy = list(map(lambda x: x["boxes"], x))
                confidences = list(map(lambda x: x["scores"], x))
                category_ids = list(map(lambda x: x["labels"], x))

                return xyxy, confidences, category_ids

        else:
            raise NotImplementedError(f"Task {self.task} is not implemented.")

        return inner

    def get_model(self) -> ModelWrapper:
        self.check_train_sanity()

        # get adapt functions
        preprocess = self.get_preprocess()
        postprocess = self.get_postprocess()

        # get model
        if self.task == TaskType.OBJECT_DETECTION:
            model = AutoModelForObjectDetection.from_pretrained(
                str(self.best_ckpt_file),
            )
        elif self.task == TaskType.CLASSIFICATION:
            model = AutoModelForImageClassification.from_pretrained(
                str(self.best_ckpt_file),
            )

        model = ModelWrapper(
            model=model.eval(),
            preprocess=preprocess,
            postprocess=postprocess,
        )

        return model



from waffle_utils.log import datetime_now
from waffle_utils.utils import type_validator

from .base_field import BaseField


class Image(BaseField):
    def __init__(
        self,
        # required
        image_id: int,
        file_name: str,
        width: int,
        height: int,
        # optional
        date_captured: str = None,
    ):

        self.image_id = image_id
        self.file_name = file_name
        self.width = width
        self.height = height
        self.date_captured = date_captured

    # properties
    @property
    def image_id(self):
        return self.__image_id

    @image_id.setter
    @type_validator(int)
    def image_id(self, v):
        if v and v < 1:
            raise ValueError("id should be greater than 0.")
        self.__image_id = v

    @property
    def file_name(self):
        return self.__file_name

    @file_name.setter
    @type_validator(str)
    def file_name(self, v):
        self.__file_name = v

    @property
    def width(self):
        return self.__width

    @width.setter
    @type_validator(int)
    def width(self, v):
        self.__width = v

    @property
    def height(self):
        return self.__height

    @height.setter
    @type_validator(int)
    def height(self, v):
        self.__height = v

    @property
    def date_captured(self):
        return self.__date_captured

    @date_captured.setter
    @type_validator(str)
    def date_captured(self, v):
        if v is None:
            self.__date_captured = datetime_now()
        else:
            self.__date_captured = v

    @classmethod
    def new(
        cls,
        image_id: int,
        file_name: str,
        width: int,
        height: int,
        date_captured: str = None,
        **kwargs,
    ) -> "Image":
        """Image Format

        Args:
            image_id (int): image id. natural number.
            file_name (str): file name. relative file path.
            width (int): image width.
            height (int): image height.
            date_captured (str): date_captured string. "%Y-%m-%d %H:%M:%S"

        Returns:
            Image: image class
        """
        return cls(image_id, file_name, width, height, date_captured)

    def to_dict(self) -> dict:
        """Get Dictionary of Category

        Returns:
            dict: annotation dictionary.
        """

        cat = {
            "image_id": self.image_id,
            "file_name": self.file_name,
            "width": self.width,
            "height": self.height,
            "date_captured": self.date_captured,
        }

        return cat



"""
Tx Model Hub
See Hub documentation for more details about usage.
"""

import warnings
from pathlib import Path
from typing import Union

import autocare_dlt
import tbparse
import torch
from autocare_dlt.core.model import build_model
from autocare_dlt.tools import train
from box import Box
from torchvision import transforms as T
from waffle_utils.file import io
from waffle_utils.utils import type_validator

from waffle_hub import TaskType
from waffle_hub.hub import Hub
from waffle_hub.hub.adapter.autocare_dlt.configs import (
    get_data_config,
    get_model_config,
)
from waffle_hub.hub.model.wrapper import ModelWrapper
from waffle_hub.schema.configs import TrainConfig
from waffle_hub.utils.callback import TrainCallback

from .config import DATA_TYPE_MAP, DEFAULT_PARAMS, MODEL_TYPES, WEIGHT_PATH


class AutocareDLTHub(Hub):
    BACKEND_NAME = "autocare_dlt"
    MODEL_TYPES = MODEL_TYPES
    MULTI_GPU_TRAIN = False
    DEFAULT_PARAMS = DEFAULT_PARAMS

    DATA_TYPE_MAP = DATA_TYPE_MAP
    WEIGHT_PATH = WEIGHT_PATH

    def __init__(
        self,
        name: str,
        task: str = None,
        model_type: str = None,
        model_size: str = None,
        categories: Union[list[dict], list] = None,
        root_dir: str = None,
        backend: str = None,
        version: str = None,
        *args,
        **kwargs,
    ):
        if backend is not None and AutocareDLTHub.BACKEND_NAME != backend:
            raise ValueError(
                f"Backend {backend} is not supported. Please use {AutocareDLTHub.BACKEND_NAME}"
            )

        if version is not None and autocare_dlt.__version__ != version:
            warnings.warn(
                f"You've loaded the Hub created with autocare_dlt=={version}, \n"
                + f"but the installed version is {autocare_dlt.__version__}."
            )

        super().__init__(
            name=name,
            backend=AutocareDLTHub.BACKEND_NAME,
            version=autocare_dlt.__version__,
            task=task,
            model_type=model_type,
            model_size=model_size,
            categories=categories,
            root_dir=root_dir,
        )

    @classmethod
    def new(
        cls,
        name: str,
        task: str = None,
        model_type: str = None,
        model_size: str = None,
        categories: Union[list[dict], list] = None,
        root_dir: str = None,
        *args,
        **kwargs,
    ):
        """Create Tx Model Hub.

        Args:
            name (str): Hub name
            task (str, optional): Task Name. See UltralyticsHub.TASKS. Defaults to None.
            model_type (str, optional): Model Type. See UltralyticsHub.MODEL_TYPES. Defaults to None.
            model_size (str, optional): Model Size. See UltralyticsHub.MODEL_SIZES. Defaults to None.
            categories (Union[list[dict], list]): class dictionary or list. [{"supercategory": "name"}, ] or ["name",].
            root_dir (str, optional): Root directory of hub repository. Defaults to None.
        """

        warnings.warn("UltralyticsHub.new() is deprecated. Please use Hub.new() instead.")

        return cls(
            name=name,
            task=task,
            model_type=model_type,
            model_size=model_size,
            categories=categories,
            root_dir=root_dir,
        )

    @property
    def categories(self) -> list[dict]:
        return self.__categories

    @categories.setter
    @type_validator(list)
    def categories(self, v):
        if isinstance(v[0], str):
            v = [{"supercategory": "object", "name": n} for n in v]
        elif isinstance(v[0], dict) and "supercategory" not in v[0]:
            # TODO: Temporal solution for DLT classification: Not supported multi-task yet.
            v_ = []
            for k, cls in v[0].items():
                for c in cls:
                    v_.append({"supercategory": k, "name": c})
            v = v_
        self.__categories = v

    # Hub Utils
    def get_preprocess(self, *args, **kwargs):

        if self.task == TaskType.OBJECT_DETECTION:
            normalize = T.Normalize([0, 0, 0], [1, 1, 1], inplace=True)

            def preprocess(x, *args, **kwargs):
                return normalize(x)

        elif self.task == TaskType.CLASSIFICATION:
            normalize = T.Normalize([0, 0, 0], [1, 1, 1], inplace=True)

            def preprocess(x, *args, **kwargs):
                return normalize(x)

        elif self.task == TaskType.TEXT_RECOGNITION:
            normalize = T.Normalize([0, 0, 0], [1, 1, 1], inplace=True)

            def preprocess(x, *args, **kwargs):
                return normalize(x)

        else:
            raise NotImplementedError(f"task {self.task} is not supported yet")

        return preprocess

    def get_postprocess(self, *args, **kwargs):

        if self.task == TaskType.OBJECT_DETECTION:

            def inner(x: torch.Tensor, *args, **kwargs):
                xyxy = x[0]
                scores = x[1]
                class_ids = x[2]

                return xyxy, scores, class_ids

        elif self.task == TaskType.CLASSIFICATION:

            def inner(x: torch.Tensor, *args, **kwargs):
                x = [t.squeeze() for t in x]
                return x

        elif self.task == TaskType.TEXT_RECOGNITION:

            def inner(x: torch.Tensor, *args, **kwargs):
                scores, character_class_ids = x.max(dim=-1)
                return character_class_ids, scores

        else:
            raise NotImplementedError(f"task {self.task} is not supported yet")

        return inner

    def get_metrics(self) -> list[dict]:
        """Get metrics from tensorboard log directory.
        Args:
            tbdir (Union[str, Path]): tensorboard log directory
        Returns:
            list[dict]: list of metrics
        """
        tb_log_dir = self.artifact_dir / "train" / "tensorboard"
        if not tb_log_dir.exists():
            return []

        sr = tbparse.SummaryReader(tb_log_dir)
        df = sr.scalars

        # Sort the data frame by step.
        # Make a list of dictionaries of tag and value.
        if df.empty:
            return []

        metrics = (
            df.sort_values("step")
            .groupby("step")
            .apply(lambda x: [{"tag": s, "value": v} for s, v in zip(x.tag, x.value)])
            .to_list()
        )

        return metrics

    # Train Hook
    def on_train_start(self, cfg: TrainConfig):
        # set data
        cfg.dataset_path: Path = Path(cfg.dataset_path)
        data_config = get_data_config(
            self.DATA_TYPE_MAP[self.task],
            cfg.image_size if isinstance(cfg.image_size, list) else [cfg.image_size, cfg.image_size],
            cfg.batch_size,
            cfg.workers,
            str(cfg.dataset_path / "train.json"),
            str(cfg.dataset_path / "images"),
            str(cfg.dataset_path / "val.json"),
            str(cfg.dataset_path / "images"),
            str(cfg.dataset_path / "test.json"),
            str(cfg.dataset_path / "images"),
        )
        if self.model_type == "LicencePlateRecognition":
            data_config["data"]["mode"] = "lpr"

        cfg.data_config = self.artifact_dir / "data.json"
        io.save_json(data_config, cfg.data_config, create_directory=True)
        categories = (
            self.categories
            if self._Hub__task == TaskType.CLASSIFICATION
            else [x["name"] for x in self.categories]
        )

        model_config = get_model_config(
            self.model_type,
            self.model_size,
            categories,
            cfg.seed,
            cfg.learning_rate,
            cfg.letter_box,
            cfg.epochs,
        )

        cfg.model_config = self.artifact_dir / "model.json"
        io.save_json(model_config, cfg.model_config, create_directory=True)

        # pretrained model
        cfg.pretrained_model = (
            cfg.pretrained_model
            if cfg.pretrained_model is not None
            else self.WEIGHT_PATH[self.task][self.model_type][self.model_size]
        )
        if not Path(cfg.pretrained_model).exists():
            cfg.pretrained_model = None
            warnings.warn(f"{cfg.pretrained_model} does not exists. Train from scratch.")

        cfg.dataset_path = str(cfg.dataset_path.absolute())

    def training(self, cfg: TrainConfig, callback: TrainCallback):
        results = train.run(
            exp_name="train",
            model_cfg=str(cfg.model_config),
            data_cfg=str(cfg.data_config),
            gpus="-1" if cfg.device == "cpu" else str(cfg.device),
            output_dir=str(self.artifact_dir),
            ckpt=cfg.pretrained_model,
            overwrite=True,
        )
        if results is None:
            raise RuntimeError("Training failed")
        del results

    def on_train_end(self, cfg: TrainConfig):
        io.copy_file(
            self.artifact_dir / "train" / "best_ckpt.pth",
            self.best_ckpt_file,
            create_directory=True,
        )
        io.copy_file(
            self.artifact_dir / "train" / "last_epoch_ckpt.pth",
            self.last_ckpt_file,
            create_directory=True,
        )
        io.save_json(self.get_metrics(), self.metric_file)

    # Inference Hook
    def get_model(self):
        """Get model.
        Returns:
            ModelWrapper: Model wrapper
        """
        self.check_train_sanity()

        # get adapt functions
        preprocess = self.get_preprocess()
        postprocess = self.get_postprocess()

        # get model
        categories = [x["name"] for x in self.categories]
        cfg = io.load_json(self.artifact_dir / "model.json")
        cfg["ckpt"] = str(self.best_ckpt_file)
        if self.task == TaskType.TEXT_RECOGNITION:
            cfg["model"]["Prediction"]["num_classes"] = len(categories) + 1
        else:
            cfg["model"]["head"]["num_classes"] = len(categories)
        cfg["num_classes"] = len(categories)
        model, categories = build_model(Box(cfg), strict=True)

        model = ModelWrapper(
            model=model.eval(),
            preprocess=preprocess,
            postprocess=postprocess,
        )

        return model



import time
from pathlib import Path

import torch

from waffle_hub import TaskType
from waffle_hub.dataset import Dataset
from waffle_hub.hub import Hub
from waffle_hub.schema.result import (
    EvaluateResult,
    ExportResult,
    InferenceResult,
    TrainResult,
)


def _train(hub, dataset: Dataset, image_size: int, hold: bool = True):
    result: TrainResult = hub.train(
        dataset=dataset,
        epochs=1,
        image_size=image_size,
        batch_size=4,
        pretrained_model=None,
        device="cpu",
        workers=0,
        hold=hold,
    )

    if not hold:
        assert hasattr(result, "callback")
        while not result.callback.is_finished() and not result.callback.is_failed():
            time.sleep(1)
        assert result.callback.is_finished()
        assert not result.callback.is_failed()

    assert len(result.metrics) >= 1
    assert len(result.eval_metrics) >= 1
    assert Path(result.best_ckpt_file).exists()
    # assert Path(result.last_ckpt_file).exists()

    return result


def _evaluate(hub, dataset: Dataset, hold: bool = True):

    result: EvaluateResult = hub.evaluate(
        dataset=dataset,
        device="cpu",
        workers=0,
        hold=hold,
    )

    if not hold:
        assert hasattr(result, "callback")
        while not result.callback.is_finished() and not result.callback.is_failed():
            time.sleep(1)
        assert result.callback.is_finished()
        assert not result.callback.is_failed()

    assert len(result.eval_metrics) >= 1

    return result


def _inference(hub, source: str, hold: bool = True):

    result: InferenceResult = hub.inference(
        source=source,
        draw=True,
        device="cpu",
        workers=0,
        hold=hold,
    )

    if not hold:
        assert hasattr(result, "callback")
        while not result.callback.is_finished() and not result.callback.is_failed():
            time.sleep(1)
        assert result.callback.is_finished()
        assert not result.callback.is_failed()

    assert len(result.predictions) >= 1
    assert Path(result.draw_dir).exists()

    return result


def _export(hub, half: bool = False, hold: bool = True):
    result: ExportResult = hub.export(
        hold=hold,
        half=half,
        device="cpu",
    )

    if not hold:
        assert hasattr(result, "callback")
        while not result.callback.is_finished() and not result.callback.is_failed():
            time.sleep(1)
        assert result.callback.is_finished()
        assert not result.callback.is_failed()

    assert Path(result.export_file).exists()

    return result


def _feature_extraction(
    hub,
    image_size: int,
):

    model = hub.get_model()

    layer_names = model.get_layer_names()
    assert len(layer_names) > 0

    x = torch.randn(1, 3, image_size, image_size)
    layer_name = layer_names[-1]
    x, feature_maps = model.get_feature_maps(x, layer_name)
    assert len(feature_maps) == 1


def _benchmark(hub, image_size):
    hub.benchmark(device="cpu", half=False, image_size=image_size)


def _util(hub):
    name = hub.name
    backend = hub.backend
    root_dir = hub.root_dir

    hub_class = Hub.get_hub_class(backend)
    assert hub_class == type(hub)

    hub_loaded = Hub.load(name, root_dir)
    assert isinstance(hub_loaded, type(hub))


def _total(hub, dataset: Dataset, image_size: int, hold: bool = True):

    _train(hub, dataset, image_size, hold=hold)
    _evaluate(hub, dataset, hold=hold)
    _inference(hub, dataset.raw_image_dir, hold=hold)
    _export(hub, half=False, hold=hold)
    # _export(hub, half=True, hold=hold)  # cpu cannot be half
    _feature_extraction(hub, image_size)
    _benchmark(hub, image_size)
    _util(hub)


def test_ultralytics_segmentation(instance_segmentation_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = instance_segmentation_dataset

    # test hub
    name = "test_seg"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.INSTANCE_SEGMENTATION,
        model_type="yolov8",
        model_size="n",
        categories=dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size)


def test_ultralytics_object_detection(object_detection_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = object_detection_dataset

    # test hub
    name = "test_det"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.OBJECT_DETECTION,
        model_type="yolov8",
        model_size="n",
        categories=dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size)


def test_ultralytics_classification(classification_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = classification_dataset

    # test hub
    name = "test_cls"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.CLASSIFICATION,
        model_type="yolov8",
        model_size="n",
        categories=classification_dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size)


def test_transformers_object_detection(object_detection_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = object_detection_dataset

    # test hub
    name = "test_det"
    hub = Hub.new(
        name=name,
        backend="transformers",
        task=TaskType.OBJECT_DETECTION,
        model_type="YOLOS",
        model_size="tiny",
        categories=object_detection_dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size)


def test_transformers_classification(classification_dataset: Dataset, tmpdir: Path):
    image_size = 224
    dataset = classification_dataset

    # test hub
    name = "test_cls"
    hub = Hub.new(
        name=name,
        backend="transformers",
        task=TaskType.CLASSIFICATION,
        model_type="ViT",
        model_size="tiny",
        categories=classification_dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size)


def test_non_hold(classification_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = classification_dataset

    # test hub
    name = "test_cls"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.CLASSIFICATION,
        model_type="yolov8",
        model_size="n",
        categories=classification_dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, hold=False)


def test_autocare_dlt_object_detection(object_detection_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = object_detection_dataset

    # test hub
    name = "test_det"
    hub = Hub.new(
        name=name,
        backend="autocare_dlt",
        task=TaskType.OBJECT_DETECTION,
        model_type="YOLOv5",
        model_size="s",
        categories=object_detection_dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size)


def test_autocare_dlt_classification(classification_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = classification_dataset

    # temporal solution
    super_cat = [[c.supercategory, c.name] for c in dataset.get_categories()]
    super_cat_dict = {}
    for super_cat, cat in super_cat:
        if super_cat not in super_cat_dict:
            super_cat_dict[super_cat] = []
        super_cat_dict[super_cat].append(cat)
    super_cat_dict_list = []

    for super_cat, cat in super_cat_dict.items():
        super_cat_dict_list.append({super_cat: cat})

    # test hub
    name = "test_cls"
    hub = Hub.new(
        name=name,
        backend="autocare_dlt",
        task=TaskType.CLASSIFICATION,
        model_type="Classifier",
        model_size="s",
        categories=super_cat_dict_list,
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size)


def test_autocare_dlt_text_recognition(text_recognition_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = text_recognition_dataset

    # test hub
    name = "test_ocr"
    hub = Hub.new(
        name=name,
        backend="autocare_dlt",
        task=TaskType.TEXT_RECOGNITION,
        model_type="TextRecognition",
        model_size="s",
        categories=dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size)



from dataclasses import dataclass

from waffle_hub.schema.base_schema import BaseSchema


@dataclass
class TrainResult(BaseSchema):
    best_ckpt_file: str = None
    last_ckpt_file: str = None
    metrics: list[list[dict]] = None
    eval_metrics: list[dict] = None


@dataclass
class EvaluateResult(BaseSchema):
    eval_metrics: list[dict] = None


@dataclass
class InferenceResult(BaseSchema):
    predictions: list[dict[list]] = None
    draw_dir: str = None


@dataclass
class ExportResult(BaseSchema):
    export_file: str = None



from pathlib import Path
from typing import Union

from waffle_utils.file import io

from waffle_hub import TaskType
from waffle_hub.utils.conversion import convert_rle_to_polygon


def _export_coco(
    self,
    export_dir: Path,
    train_ids: list,
    val_ids: list,
    test_ids: list,
    unlabeled_ids: list,
):
    """Export dataset to COCO format

    Args:
        export_dir (Path): Path to export directory
        train_ids (list): List of train ids
        val_ids (list): List of validation ids
        test_ids (list): List of test ids
        unlabeled_ids (list): List of unlabeled ids
    """
    io.make_directory(export_dir)

    image_dir = export_dir / "images"

    for split, image_ids in zip(
        ["train", "val", "test", "unlabeled"],
        [train_ids, val_ids, test_ids, unlabeled_ids],
    ):
        if len(image_ids) == 0:
            continue

        coco = {
            "categories": [
                {
                    "id": category.category_id,
                    "name": category.name,
                    "supercategory": category.supercategory,
                }
                for category in self.get_categories()
            ],
            "images": [],
            "annotations": [],
        }

        for image in self.get_images(image_ids):
            image_path = self.raw_image_dir / image.file_name
            image_dst_path = image_dir / image.file_name
            io.copy_file(image_path, image_dst_path, create_directory=True)

            d = image.to_dict()
            image_id = d.pop("image_id")
            coco["images"].append({"id": image_id, **d})

            annotations = self.get_annotations(image_id)
            for annotation in annotations:
                d = annotation.to_dict()
                if d.get("segmentation", None):
                    if isinstance(d["segmentation"], dict):
                        d["segmentation"] = convert_rle_to_polygon(d["segmentation"])
                annotation_id = d.pop("annotation_id")
                coco["annotations"].append({"id": annotation_id, **d})

        io.save_json(coco, export_dir / f"{split}.json", create_directory=True)


def export_coco(self, export_dir: Union[str, Path]) -> str:
    """Export dataset to COCO format

    Args:
        export_dir (str): Path to export directory

    Returns:
        str: Path to export directory
    """
    export_dir = Path(export_dir)

    train_ids, val_ids, test_ids, _ = self.get_split_ids()

    if self.task == TaskType.CLASSIFICATION:
        _export_coco(self, export_dir, train_ids, val_ids, test_ids, [])
    elif self.task == TaskType.OBJECT_DETECTION:
        _export_coco(self, export_dir, train_ids, val_ids, test_ids, [])
    elif self.task == TaskType.INSTANCE_SEGMENTATION:
        _export_coco(self, export_dir, train_ids, val_ids, test_ids, [])
    else:
        raise ValueError(f"Unsupported task type: {self.task}")

    return str(export_dir)



"""_summary_

Raises:
    NotImplementedError: _description_
    ValueError: _description_

Returns:
    _type_: _description_
"""

from typing import Union

import numpy as np
import torch
import torch.nn.functional as F
from torchvision.ops import batched_nms

from waffle_hub import TaskType
from waffle_hub.schema.data import ImageInfo
from waffle_hub.schema.fields import Annotation
from waffle_hub.utils.conversion import convert_mask_to_polygon


class PreprocessFunction:
    pass


class PostprocessFunction:
    pass


class ResultParser:
    pass


class ClassificationResultParser(ResultParser):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(
        self, results: list[torch.Tensor], image_infos: list[ImageInfo] = None, *args, **kwargs
    ) -> list[Annotation]:
        parseds = []

        results = results[0]  # TODO: multi label
        scores, class_ids = results.cpu().topk(results.shape[1], dim=-1)
        results = torch.cat([class_ids.unsqueeze(-1), scores.unsqueeze(-1)], dim=-1)
        for result in results:
            parsed = []
            for class_id, score in result:
                parsed.append(
                    Annotation.classification(category_id=int(class_id) + 1, score=float(score))
                )
            parseds.append(parsed)
        return parseds


class ObjectDetectionResultParser(ResultParser):
    def __init__(
        self, confidence_threshold: float = 0.25, iou_threshold: float = 0.5, *args, **kwargs
    ):
        self.confidence_threshold = confidence_threshold
        self.iou_threshold = iou_threshold

    def __call__(
        self, results: list[torch.Tensor], image_infos: list[ImageInfo], *args, **kwargs
    ) -> list[Annotation]:
        parseds = []

        bboxes_batch, confs_batch, class_ids_batch = results
        for bboxes, confs, class_ids, image_info in zip(
            bboxes_batch, confs_batch, class_ids_batch, image_infos
        ):

            mask = confs > self.confidence_threshold

            bboxes, confs, class_ids = (
                bboxes[mask],
                confs[mask],
                class_ids[mask],
            )
            idxs = batched_nms(bboxes, confs, class_ids, self.iou_threshold)

            bboxes = bboxes[idxs, :].cpu()
            confs = confs[idxs].cpu()
            class_ids = class_ids[idxs].cpu()

            W, H = image_info.input_shape
            left_pad, top_pad = image_info.pad
            ori_w, ori_h = image_info.ori_shape
            new_w, new_h = image_info.new_shape

            parsed = []
            for (x1, y1, x2, y2), conf, class_id in zip(bboxes, confs, class_ids):

                x1 = max(float((x1 * W - left_pad) / new_w * ori_w), 0)
                y1 = max(float((y1 * H - top_pad) / new_h * ori_h), 0)
                x2 = min(float((x2 * W - left_pad) / new_w * ori_w), ori_w)
                y2 = min(float((y2 * H - top_pad) / new_h * ori_h), ori_h)

                parsed.append(
                    Annotation.object_detection(
                        category_id=int(class_id) + 1,
                        bbox=[x1, y1, x2 - x1, y2 - y1],
                        area=float((x2 - x1) * (y2 - y1)),
                        score=float(conf),
                    )
                )
            parseds.append(parsed)
        return parseds


class InstanceSegmentationResultParser(ObjectDetectionResultParser):
    def __init__(
        self, confidence_threshold: float = 0.25, iou_threshold: float = 0.5, *args, **kwargs
    ):
        self.confidence_threshold = confidence_threshold
        self.iou_threshold = iou_threshold
        super().__init__(confidence_threshold, iou_threshold, *args, **kwargs)

    def __call__(
        self, results: list[torch.Tensor], image_infos: list[ImageInfo], *args, **kwargs
    ) -> list[Annotation]:
        parseds = []
        bboxes_batch, confs_batch, class_ids_batch, masks_batch = results
        for bboxes, confs, class_ids, masks, image_info in zip(
            bboxes_batch,
            confs_batch,
            class_ids_batch,
            masks_batch,
            image_infos,
        ):

            mask = confs > self.confidence_threshold

            bboxes, confs, class_ids, masks = (
                bboxes[mask],
                confs[mask],
                class_ids[mask],
                masks[mask],
            )
            idxs = batched_nms(bboxes, confs, class_ids, self.iou_threshold)

            bboxes = bboxes[idxs, :].cpu()
            confs = confs[idxs].cpu()
            class_ids = class_ids[idxs].cpu()

            masks = masks[idxs, :]
            masks = (
                F.interpolate(
                    input=masks.gt_(0.5).unsqueeze(1),
                    size=image_info.ori_shape,
                    mode="bilinear",
                    align_corners=False,
                )
                .cpu()
                .squeeze(1)
            )

            W, H = image_info.input_shape
            left_pad, top_pad = image_info.pad
            ori_w, ori_h = image_info.ori_shape
            new_w, new_h = image_info.new_shape

            parsed = []
            for (x1, y1, x2, y2), conf, class_id, mask in zip(bboxes, confs, class_ids, masks):

                x1 = max(float((x1 * W - left_pad) / new_w * ori_w), 0)
                y1 = max(float((y1 * H - top_pad) / new_h * ori_h), 0)
                x2 = min(float((x2 * W - left_pad) / new_w * ori_w), ori_w)
                y2 = min(float((y2 * H - top_pad) / new_h * ori_h), ori_h)

                # clean non roi area
                mask[:, : round(x1)] = 0
                mask[:, round(x2) :] = 0
                mask[: round(y1), :] = 0
                mask[round(y2) :, :] = 0

                segment = convert_mask_to_polygon(mask.numpy().astype(np.uint8))

                parsed.append(
                    Annotation.instance_segmentation(
                        category_id=int(class_id) + 1,
                        bbox=[x1, y1, x2 - x1, y2 - y1],
                        area=float((x2 - x1) * (y2 - y1)),
                        score=float(conf),
                        segmentation=segment,
                    )
                )
            parseds.append(parsed)
        return parseds


class TextRecognitionResultParser(ResultParser):
    def __init__(self, categories, *args, **kwargs):
        self.categories = categories
        self.category_names = [""] + [d["name"] for d in self.categories]

    def __call__(
        self, results: list[torch.Tensor], image_infos: list[ImageInfo] = None, *args, **kwargs
    ) -> list[Annotation]:
        parseds = []

        pred_batch, conf_batch = results
        for preds, confs in zip(pred_batch, conf_batch):
            mask = preds > 0
            preds, confs = preds[mask], confs[mask]

            text = "".join([self.category_names[pred] for pred in preds])

            parsed = []
            parsed.append(
                Annotation.text_recognition(
                    score=list(map(float, confs)),
                    caption=text,
                )
            )
            parseds.append(parsed)

        return parseds


def get_parser(task: str):
    if task == TaskType.CLASSIFICATION:
        return ClassificationResultParser
    elif task == TaskType.OBJECT_DETECTION:
        return ObjectDetectionResultParser
    elif task == TaskType.INSTANCE_SEGMENTATION:
        return InstanceSegmentationResultParser
    elif task == TaskType.TEXT_RECOGNITION:
        return TextRecognitionResultParser


class ModelWrapper(torch.nn.Module):
    def __init__(
        self,
        model: torch.nn.Module,
        preprocess: PreprocessFunction,
        postprocess: PostprocessFunction,
    ):
        """
        Model Wrapper.
        Use this wrapper when inference, export.

        Args:
            model (torch.nn.Module): model
            preprocess (PreprocessFunction):
                Preprocess Function that
                recieves [batch, channel, height, width] (0~1),
                and
                outputs [batch, channel, height, width].

            postprocess (PostprocessFunction):
                Postprocess Function that
                recieves model raw output,
                and
                outputs results that fit with our convention.

                Classification:
                    [
                        [batch, class_num],
                    ]  # scores per attribute
                Detection:
                    [
                        [batch, bbox_num, 4(x1, y1, x2, y2)],  # bounding box
                        [batch, bbox_num],  # confidence
                        [batch, bbox_num],  # class id
                    ]
                Segmentation:
                    [
                        [batch, bbox_num, 4(x1, y1, x2, y2)],  # bounding box
                        [batch, bbox_num],  # confidence
                        [batch, bbox_num],  # class id
                        [batch, mask(H, W)] # warning: mask size and image size are not same
                    ]
        """
        super().__init__()
        self.model = model
        self.preprocess = preprocess
        self.postprocess = postprocess

    def forward(self, x):
        _, _, H, W = x.shape
        x = self.preprocess(x)
        x = self.model(x)
        x = self.postprocess(x, image_size=(W, H))
        return x

    def get_layer_names(self) -> list[str]:
        """
        Get all layer names in model.
        """
        return [name for name, _ in self.model.named_modules()]

    def get_layers(self, layer_names: Union[list[str], str]) -> list[torch.nn.Module]:
        """
        Get layer in model by name.

        Args:
            layer_names (Union[list[str], str]): layer names to get

        Returns:
            layers (list[torch.nn.Module]): layers
        """
        if isinstance(layer_names, str):
            layer_names = [layer_names]
        return [layer for name, layer in self.model.named_modules() if name in layer_names]

    def _convert_to_feature_map(self, feature: torch.Tensor) -> torch.Tensor:
        """
        Convert feature to feature map.
        """
        if len(feature.shape) == 4:  # Convolution Feature Map
            return feature
        elif len(feature.shape) == 3:  # Linear(ViT) Feature Map
            batch, dim, channel = feature.shape
            size = int(np.sqrt(dim - 1))
            return feature[:, 1:].view(
                batch, channel, size, size
            )  # TODO: first token is cls token (usually)
        else:
            raise ValueError(f"Unsupported feature map type. {feature.shape}")

    def get_feature_maps(
        self, x, layer_names: Union[list[str], str] = None
    ) -> tuple[torch.Tensor, dict]:
        """
        Get feature maps from model.

        Args:
            x (torch.Tensor): input image
            layer_names (Union[list[str], str]): layer names to get feature maps

        Returns:
            x (torch.Tensor): model output
            feature_maps (dict): feature maps
        """

        feature_maps = {}

        def hook(name):
            def hook_fn(m, i, o):
                feature_maps[name] = o

            return hook_fn

        if layer_names is None:
            layer_names = self.get_layer_names()[-1]
        elif isinstance(layer_names, str):
            layer_names = [layer_names]

        for name, module in self.model.named_modules():
            if name in layer_names:
                module.register_forward_hook(hook(name))

        x = self.forward(x)

        return x, feature_maps

    def get_cam(
        self,
        x: torch.Tensor,
        layer_name: str,
        class_id: int = None,
    ) -> torch.Tensor:
        """
        TODO: experimental
        Get class activation map.

        Args:
            x (torch.Tensor): input image
            layer_name (str): layer name to get feature maps
            class_id (int): class id to get cam
            image_size (tuple[int, int]): image size
            upsample_size (tuple[int, int]): upsample size

        Returns:
            cam (torch.Tensor): class activation map
        """

        activation_list = []
        gradient_list = []

        def hook(name):
            def hook_fn(m, i, o):
                activation_list.append(self._convert_to_feature_map(o).cpu().detach())

                if not hasattr(o, "requires_grad" or not o.requires_grad):
                    return

                def _store_grad(grad):
                    gradient_list.insert(0, self._convert_to_feature_map(grad).cpu().detach())

                o.register_hook(_store_grad)

            return hook_fn

        layer = self.get_layers(layer_name)[0]
        layer.requires_grad_(True)
        handle = layer.register_forward_hook(hook(layer_name))

        self.model.zero_grad()
        output = self.forward(x)[0]
        if class_id is None:
            class_id = torch.argmax(output[0])
        output[:, class_id].backward()

        activations = activation_list[0]
        gradients = gradient_list[0]

        if activations.shape != gradients.shape:
            raise ValueError(
                f"Activation shape {activations.shape} and gradient shape {gradients.shape} are not same."
            )

        weights = torch.mean(gradients, dim=(2, 3), keepdim=True)
        cam = torch.sum(weights * activations, dim=1, keepdim=True)

        # release hooks
        handle.remove()

        # normalize cam to (0, 1)
        cam = torch.nn.functional.relu(cam)
        cam_max = torch.max(cam)
        cam_min = torch.min(cam)
        cam = (cam - cam_min) / (cam_max - cam_min)

        # resize cam
        cam = torch.nn.functional.interpolate(
            cam, size=x.shape[-2:], mode="bilinear", align_corners=False
        )

        return cam



import logging
from pathlib import Path
from typing import Union

import cv2
import numpy as np
from PIL import Image, ImageDraw, ImageFont
from waffle_utils.file.network import get_file_from_url

from waffle_hub import TaskType
from waffle_hub.schema.fields import Annotation

FONT_FACE = cv2.FONT_HERSHEY_DUPLEX
FONT_SCALE = 1.0
FONT_WEIGHT = 2
THICKNESS = 2
FONT_URL = "https://raw.githubusercontent.com/snuailab/assets/main/waffle/fonts/gulim.ttc"
FONT_NAME = "gulim.ttc"


# random colors with 1000 categories
colors = np.random.randint(0, 255, (1000, 3), dtype="uint8").tolist()


def draw_classification(
    image: np.ndarray,
    annotation: Annotation,
    names: list[str],
    loc_x: int = 10,
    loc_y: int = 30,
):
    category_id: int = annotation.category_id
    score: float = annotation.score

    image = cv2.putText(
        image,
        f"{names[category_id-1]}" + (f": {score:.2f}" if score else ""),
        (loc_x, loc_y),
        FONT_FACE,
        FONT_SCALE,
        colors[category_id - 1],
        FONT_WEIGHT,
    )
    return image


def draw_object_detection(
    image: np.ndarray,
    annotation: Annotation,
    names: list[str],
    score: float = None,
):
    x1, y1, w, h = annotation.bbox
    x2 = x1 + w
    y2 = y1 + h

    category_id: int = annotation.category_id
    score: float = annotation.score

    image = cv2.putText(
        image,
        f"{names[category_id-1]}" + (f": {score:.2f}" if score else ""),
        (int(x1), int(y1) - 3),
        FONT_FACE,
        FONT_SCALE,
        colors[category_id - 1],
        FONT_WEIGHT,
    )
    image = cv2.rectangle(
        image,
        (int(x1), int(y1)),
        (int(x2), int(y2)),
        colors[category_id - 1],
        THICKNESS,
    )
    return image


def draw_instance_segmentation(
    image: np.ndarray,
    annotation: Annotation,
    names: list[str],
    score: float = None,
):
    image = draw_object_detection(image, annotation, names, score)
    segments: list = annotation.segmentation

    if len(segments) == 0:
        return image

    alpha = np.zeros_like(image)
    for segment in segments:
        segment = np.array(segment).reshape(-1, 2).astype(int)
        alpha = cv2.fillPoly(
            alpha,
            [segment],
            colors[annotation.category_id - 1],
        )
    mask = alpha > 0
    image[mask] = cv2.addWeighted(alpha, 0.3, image, 0.7, 0)[mask]

    return image


def draw_text_recognition(
    image: np.ndarray,
    annotation: Annotation,
    loc_x: int = 0,
    loc_y: int = 10,
):
    # download font
    try:
        global FONT_NAME
        if not Path(FONT_NAME).exists():
            get_file_from_url(FONT_URL, FONT_NAME, True)
        font = ImageFont.truetype(FONT_NAME, int(FONT_SCALE) * 25)
    except:
        font = ImageFont.load_default()
        logging.warning("Don't load font file, Using default font.")

    img_pil = Image.fromarray(image)
    draw = ImageDraw.Draw(img_pil)
    draw.text(
        (loc_x, loc_y),
        annotation.caption,
        font=font,
        fill=tuple(colors[0]),
        stroke_width=FONT_WEIGHT,
    )

    image = np.array(img_pil)

    return image


def draw_results(
    image: Union[np.ndarray, str],
    results: list[Annotation],
    names: list[str],
):

    if isinstance(image, str):
        image = cv2.imread(image)

    task_results = {task: [] for task in TaskType}
    for result in results:
        task_results[result.task.upper()].append(result)

    for i, result in enumerate(task_results[TaskType.CLASSIFICATION], start=1):
        image = draw_classification(
            image,
            result,
            names=names,
            loc_x=10,
            loc_y=30 * i,
        )

    for i, result in enumerate(task_results[TaskType.OBJECT_DETECTION], start=1):
        image = draw_object_detection(image, result, names=names)

    for i, result in enumerate(task_results[TaskType.INSTANCE_SEGMENTATION], start=1):
        image = draw_instance_segmentation(image, result, names=names)

    for i, result in enumerate(task_results[TaskType.TEXT_RECOGNITION], start=1):
        image = draw_text_recognition(image, result, loc_x=10, loc_y=10)

    return image



import time
from pathlib import Path

from waffle_hub import TaskType
from waffle_hub.dataset import Dataset
from waffle_hub.hub import Hub
from waffle_hub.schema.result import TrainResult


def _train(hub, dataset: Dataset, image_size: int, hold: bool = True):
    result: TrainResult = hub.train(
        dataset=dataset,
        epochs=1,
        image_size=image_size,
        batch_size=4,
        pretrained_model=None,
        device="0,1",
        workers=0,
        hold=hold,
    )

    if not hold:
        assert hasattr(result, "callback")
        while not result.callback.is_finished() and not result.callback.is_failed():
            time.sleep(1)
        assert result.callback.is_finished()
        assert not result.callback.is_failed()

    print(hub.metric_file, result.metrics)
    assert len(result.metrics) >= 1
    assert Path(result.best_ckpt_file).exists()
    # assert Path(result.last_ckpt_file).exists()

    return result


def test_ultralytics_segmentation(instance_segmentation_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = instance_segmentation_dataset

    # test hub
    name = "test_seg_ultralytics"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.INSTANCE_SEGMENTATION,
        model_type="yolov8",
        model_size="n",
        categories=dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _train(hub, dataset, image_size)


def test_ultralytics_object_detection(object_detection_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = object_detection_dataset

    # test hub
    name = "test_det_ultralytics"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.OBJECT_DETECTION,
        model_type="yolov8",
        model_size="n",
        categories=dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _train(hub, dataset, image_size)


def test_ultralytics_classification(classification_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = classification_dataset

    # test hub
    name = "test_cls_ultralytics"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.CLASSIFICATION,
        model_type="yolov8",
        model_size="n",
        categories=classification_dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _train(hub, dataset, image_size)


# def test_transformers_classification(classification_dataset: Dataset, tmpdir: Path):
#     image_size = 224
#     dataset = classification_dataset

#     # test hub
#     name = "test_cls_transformers"
#     hub = Hub.new(
#         name=name,
#         backend="transformers",
#         task=TaskType.CLASSIFICATION,
#         model_type="ViT",
#         model_size="tiny",
#         categories=classification_dataset.get_category_names(),
#         root_dir=tmpdir,
#     )
#     hub = Hub.load(name=name, root_dir=tmpdir)
#     hub: Hub = Hub.from_model_config(
#         name=name + "_from_model_config",
#         model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
#         root_dir=tmpdir,
#     )

#     _train(hub, dataset, image_size)


# def test_transformers_object_detection(object_detection_dataset: Dataset, tmpdir: Path):
#     image_size = 32
#     dataset = object_detection_dataset

#     # test hub
#     name = "test_det_transformers"
#     hub = Hub.new(
#         name=name,
#         backend="transformers",
#         task=TaskType.OBJECT_DETECTION,
#         model_type="YOLOS",
#         model_size="tiny",
#         categories=dataset.get_category_names(),
#         root_dir=tmpdir,
#     )
#     hub = Hub.load(name=name, root_dir=tmpdir)
#     hub: Hub = Hub.from_model_config(
#         name=name + "_from_model_config",
#         model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
#         root_dir=tmpdir,
#     )

#     _train(hub, dataset, image_size)



import inspect

import fire

from waffle_hub.dataset import Dataset
from waffle_hub.utils.base_cli import BaseCLI, cli


class DatasetInstance(BaseCLI):
    def __init__(self, name: str, root_dir: str = None):
        self.dataset = None
        if name in Dataset.get_dataset_list(root_dir):
            self.dataset = Dataset.load(name, root_dir=root_dir)
        else:
            raise ValueError(f"Dataset {name} does not exist.")

        super().__init__()

    def get_object(self):
        return self.dataset


def main():
    fire.Fire(cli(Dataset, DatasetInstance), serialize=str)


if __name__ == "__main__":
    main()



import warnings
from pathlib import Path
from typing import Union

from waffle_utils.file import io

from waffle_hub import TaskType
from waffle_hub.schema.fields.image import Image
from waffle_hub.utils.conversion import merge_multi_segment


def _check_valid_file_paths(images: list[Image]) -> bool:
    """Check file paths are valid
    If the file name includes the words "images" or "labels," an error occurs during training

    Args:
        images (list[Image]): List of Image

    Returns:
        bool: True if valid
    """
    for image in images:
        file_path = Path(image.file_name)
        if "images" in file_path.parts:
            raise ValueError(
                f"The file path '{file_path}' is not allowed. Please choose a file path that does not contain the word 'images'"
            )
        if "labels" in file_path.parts:
            raise ValueError(
                f"The file path '{file_path}' is not allowed. Please choose a file path that does not contain the word 'labels'"
            )
    else:
        return True


def _export_yolo_classification(
    self,
    export_dir: Path,
    train_ids: list,
    val_ids: list,
    test_ids: list,
    unlabeled_ids: list,
):
    """Export dataset to YOLO format for classification task

    Args:
        export_dir (Path): Path to export directory
        train_ids (list): List of train ids
        val_ids (list): List of validation ids
        test_ids (list): List of test ids
        unlabeled_ids (list): List of unlabeled ids
    """
    io.make_directory(export_dir)

    for split, image_ids in zip(
        ["train", "val", "test", "unlabeled"],
        [train_ids, val_ids, test_ids, unlabeled_ids],
    ):
        if len(image_ids) == 0:
            continue

        split_dir = export_dir / split
        io.make_directory(split_dir)

        category_names = {c.category_id: c.name for c in self.get_categories()}
        for image in self.get_images(image_ids):
            image_path = self.raw_image_dir / image.file_name

            annotations = self.get_annotations(image.image_id)
            if len(annotations) > 1:
                warnings.warn(f"Multi label does not support yet. Skipping {image_path}.")
                continue
            category_id = annotations[0].category_id

            image_dst_path = split_dir / category_names[category_id] / image.file_name
            io.copy_file(image_path, image_dst_path, create_directory=True)


def _export_yolo_detection(
    self,
    export_dir: Path,
    train_ids: list,
    val_ids: list,
    test_ids: list,
    unlabeled_ids: list,
):
    """Export dataset to YOLO format for detection task

    Args:
        export_dir (Path): Path to export directory
        train_ids (list): List of train ids
        val_ids (list): List of validation ids
        test_ids (list): List of test ids
        unlabeled_ids (list): List of unlabeled ids
    """
    io.make_directory(export_dir)

    for split, image_ids in zip(
        ["train", "val", "test", "unlabeled"],
        [train_ids, val_ids, test_ids, unlabeled_ids],
    ):
        if len(image_ids) == 0:
            continue

        image_dir = export_dir / split / "images"
        label_dir = export_dir / split / "labels"

        io.make_directory(image_dir)
        io.make_directory(label_dir)

        for image in self.get_images(image_ids):
            image_path = self.raw_image_dir / image.file_name
            image_dst_path = image_dir / image.file_name
            label_dst_path = (label_dir / image.file_name).with_suffix(".txt")
            io.copy_file(image_path, image_dst_path, create_directory=True)

            W = image.width
            H = image.height

            annotations = self.get_annotations(image.image_id)
            label_txts = []
            for annotation in annotations:
                x1, y1, w, h = annotation.bbox
                x1, w = x1 / W, w / W
                y1, h = y1 / H, h / H
                cx, cy = x1 + w / 2, y1 + h / 2

                category_id = annotation.category_id - 1

                label_txts.append(f"{category_id} {cx} {cy} {w} {h}")

            io.make_directory(label_dst_path.parent)
            with open(label_dst_path, "w") as f:
                f.write("\n".join(label_txts))


def _export_yolo_segmentation(
    self,
    export_dir: Path,
    train_ids: list,
    val_ids: list,
    test_ids: list,
    unlabeled_ids: list,
):
    io.make_directory(export_dir)

    for split, image_ids in zip(
        ["train", "val", "test", "unlabeled"],
        [train_ids, val_ids, test_ids, unlabeled_ids],
    ):
        if len(image_ids) == 0:
            continue

        image_dir = export_dir / split / "images"
        label_dir = export_dir / split / "labels"

        io.make_directory(image_dir)
        io.make_directory(label_dir)

        for image in self.get_images(image_ids):
            image_path = self.raw_image_dir / image.file_name
            image_dst_path = image_dir / image.file_name
            label_dst_path = (label_dir / image.file_name).with_suffix(".txt")
            io.copy_file(image_path, image_dst_path, create_directory=True)

            W = image.width
            H = image.height

            annotations = self.get_annotations(image.image_id)
            label_txts = []
            for annotation in annotations:
                x1, y1, w, h = annotation.bbox
                x1, w = x1 / W, w / W
                y1, h = y1 / H, h / H

                category_id = annotation.category_id - 1

                segment = merge_multi_segment(annotation.segmentation, (W, H))
                segment[0::2] = [x / W for x in segment[0::2]]
                segment[1::2] = [y / H for y in segment[1::2]]
                segment = " ".join(map(str, segment))

                label_txts.append(f"{category_id} {segment}")

            io.make_directory(label_dst_path.parent)
            with open(label_dst_path, "w") as f:
                f.write("\n".join(label_txts))


def export_yolo(self, export_dir: Union[str, Path]) -> str:
    """Export dataset to YOLO format

    Args:
        export_dir (Union[str, Path]): Path to export directory

    Returns:
        str: Path to export directory
    """
    _check_valid_file_paths(self.get_images())

    export_dir = Path(export_dir)

    train_ids, val_ids, test_ids, _ = self.get_split_ids()

    if self.task == TaskType.CLASSIFICATION:
        _export_yolo_classification(self, export_dir, train_ids, val_ids, test_ids, [])
    elif self.task == TaskType.OBJECT_DETECTION:
        _export_yolo_detection(self, export_dir, train_ids, val_ids, test_ids, [])
    elif self.task == TaskType.INSTANCE_SEGMENTATION:
        _export_yolo_segmentation(self, export_dir, train_ids, val_ids, test_ids, [])
    else:
        raise ValueError(f"Unsupported task type: {self.task}")

    io.save_yaml(
        {
            "path": str(export_dir.absolute()),
            "train": "train",
            "val": "val",
            "test": "test",
            "names": {category.category_id - 1: category.name for category in self.get_categories()},
        },
        export_dir / "data.yaml",
    )

    return str(export_dir)



from pathlib import Path
from typing import Union

import cv2
import numpy as np
import torch
from torchvision import transforms as T
from waffle_utils.file import io

from waffle_hub.dataset import Dataset
from waffle_hub.schema.data import ImageInfo
from waffle_hub.schema.fields import Annotation, Category, Image


def get_images(d, recursive: bool = True) -> list[str]:
    exp = "**/*" if recursive else "*"
    image_paths = []
    for ext in ["png", "jpg", "jpeg", "bmp", "tif", "tiff"]:
        image_paths += list(Path(d).glob(exp.lower() + "." + ext))
        image_paths += list(Path(d).glob(exp.upper() + "." + ext))
    return list(
        set(
            map(
                str,
                image_paths,
            )
        )
    )


def resize_image(
    image: np.ndarray, image_size: list[int], letter_box: bool = False
) -> list[np.ndarray, ImageInfo]:
    """Resize Image.

    Args:
        image (np.ndarray): opencv image.
        image_size (list[int]): image [width, height].
        letter_box (bool): letter box.

    Returns:
        list[np.ndarray, ImageInfo]: resized image, image info.
    """

    h, w = image.shape[:2]
    W, H = image_size

    if letter_box:
        if w > h:
            ratio = W / w
            w_ = W
            h_ = round(h * ratio)

            total_pad = w_ - h_
            top = total_pad // 2
            bottom = total_pad - top
            left, right = 0, 0

        else:
            ratio = H / h
            w_ = round(w * ratio)
            h_ = H

            total_pad = h_ - w_
            left = total_pad // 2
            right = total_pad - left
            top, bottom = 0, 0

        image = cv2.resize(image, (w_, h_), interpolation=cv2.INTER_CUBIC)
        image = cv2.copyMakeBorder(image, top, bottom, left, right, None, value=(114, 114, 114))

    else:
        w_, h_ = W, H
        left, top = 0, 0
        image = cv2.resize(image, (w_, h_), interpolation=cv2.INTER_CUBIC)

    return image, ImageInfo(
        ori_shape=(w, h),
        new_shape=(w_, h_),
        input_shape=(W, H),
        pad=(left, top),
    )


def get_image_transform(image_size: Union[int, list[int]], letter_box: bool = False):

    if isinstance(image_size, int):
        image_size = [image_size, image_size]

    def transform(image: Union[np.ndarray, str]) -> tuple[torch.Tensor, ImageInfo]:
        if isinstance(image, str):
            image = cv2.imread(image)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image, image_info = resize_image(image, image_size, letter_box)
        return T.ToTensor()(image), image_info

    return transform


class ImageDataset:
    def __init__(
        self,
        image_dir: str,
        image_size: Union[int, list[int]],
        letter_box: bool = False,
    ):
        self.image_dir = image_dir
        if Path(self.image_dir).is_file():
            self.image_paths = [self.image_dir]
        else:
            self.image_paths = get_images(self.image_dir)

        self.transform = get_image_transform(image_size, letter_box)

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        image_path = self.image_paths[idx]

        image, image_info = self.transform(image_path)
        image_info.image_path = image_path

        return image, image_info

    def collate_fn(self, batch):
        images, infos = list(zip(*batch))
        return torch.stack(images, dim=0), infos

    def get_dataloader(self, batch_size: int = 4, num_workers: int = 0):
        return torch.utils.data.DataLoader(
            self,
            batch_size,
            num_workers=num_workers,
            collate_fn=self.collate_fn,
            shuffle=False,
            drop_last=False,
        )


class LabeledDataset:
    def __init__(
        self,
        dataset: Dataset,
        image_size: Union[int, list[int]],
        letter_box: bool = False,
        set_name: str = None,
    ):
        self.dataset = dataset
        self.image_dir = dataset.raw_image_dir
        self.set_name = set_name

        if self.set_name == "train":
            set_file = self.dataset.train_set_file
        elif self.set_name == "val":
            set_file = self.dataset.val_set_file
        elif self.set_name == "test":
            set_file = self.dataset.test_set_file
        else:
            set_file = None

        self.images: list[Image] = self.dataset.get_images(
            io.load_json(set_file) if set_file else None
        )
        self.image_to_annotations: dict[int, list[Annotation]] = {
            image.image_id: self.dataset.get_annotations(image.image_id) for image in self.images
        }

        self.transform = get_image_transform(image_size, letter_box)

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        image = self.images[idx]
        image_path = str(self.image_dir / image.file_name)
        annotations: list[Annotation] = self.image_to_annotations[image.image_id]

        image, image_info = self.transform(image_path)
        image_info.image_path = image_path

        return image, image_info, annotations

    def collate_fn(self, batch):
        images, infos, annotations = list(zip(*batch))
        return torch.stack(images, dim=0), infos, annotations

    def get_dataloader(self, batch_size: int = 4, num_workers: int = 0):
        return torch.utils.data.DataLoader(
            self,
            batch_size,
            num_workers=num_workers,
            collate_fn=self.collate_fn,
            shuffle=False,
            drop_last=False,
        )



from functools import reduce
from operator import eq
from typing import Union

import torch
from torchmetrics.classification import Accuracy
from torchmetrics.detection import mean_ap
from torchmetrics.text import CharErrorRate

from waffle_hub import TaskType
from waffle_hub.schema.evaluate import (
    ClassificationMetric,
    InstanceSegmentationMetric,
    ObjectDetectionMetric,
    TextRecognitionMetric,
)
from waffle_hub.schema.fields import Annotation


def convert_to_torchmetric_format(total: list[Annotation], task: TaskType, prediction: bool = False):

    datas = []
    for annotations in total:

        if task == TaskType.CLASSIFICATION:  # single attribute
            datas.append(annotations[0].category_id - 1)

        elif task == TaskType.OBJECT_DETECTION:
            data = {
                "boxes": [],
                "labels": [],
            }
            if prediction:
                data["scores"] = []

            for annotation in annotations:
                data["boxes"].append(annotation.bbox)
                data["labels"].append(annotation.category_id - 1)
                if prediction:
                    data["scores"].append(annotation.score)

            datas.append(data)

        elif task == TaskType.INSTANCE_SEGMENTATION:
            data = {
                "boxes": [],
                "labels": [],
            }
            if prediction:
                data["scores"] = []

            for annotation in annotations:
                data["boxes"].append(annotation.bbox)
                data["labels"].append(annotation.category_id - 1)
                if prediction:
                    data["scores"].append(annotation.score)

            datas.append(data)

        elif task == TaskType.TEXT_RECOGNITION:
            datas.append(annotations[0].caption)

        else:
            raise NotImplementedError

    if isinstance(datas[0], dict):
        datas = [{k: torch.tensor(v) for k, v in data.items()} for data in datas]
    elif isinstance(datas[0], int):
        datas = torch.tensor(datas)
    elif isinstance(datas[0], str):
        pass
    else:
        raise NotImplementedError

    return datas


def evaluate_classification(
    preds: list[Annotation], labels: list[Annotation], num_classes: int
) -> ClassificationMetric:

    acc = Accuracy(task="multiclass", num_classes=num_classes)(preds, labels)

    return ClassificationMetric(float(acc))


def evaluate_object_detection(
    preds: list[Annotation], labels: list[Annotation], num_classes: int
) -> ObjectDetectionMetric:

    map_dict = mean_ap.MeanAveragePrecision(
        box_format="xywh",
        iou_type="bbox",
        class_metrics=True,
        num_classes=num_classes,
    )(preds, labels)

    return ObjectDetectionMetric(float(map_dict["map"]))


def evaluate_segmentation(
    preds: list[Annotation], labels: list[Annotation], num_classes: int
) -> InstanceSegmentationMetric:

    map_dict = mean_ap.MeanAveragePrecision(
        box_format="xywh",
        iou_type="bbox",
        class_metrics=True,
        num_classes=num_classes,
    )(preds, labels)

    return InstanceSegmentationMetric(float(map_dict["map"]))


def evalute_text_recognition(
    preds: list[Annotation], labels: list[Annotation], num_classes: int
) -> ObjectDetectionMetric:

    correct = reduce(lambda n, pair: n + eq(*pair), zip(preds, labels), 0)
    acc = correct / len(preds)

    return TextRecognitionMetric(float(acc))


def evaluate_function(
    preds: list[Annotation],
    labels: list[Annotation],
    task: str,
    num_classes: int = None,
    *args,
    **kwargs
) -> Union[
    ClassificationMetric, ObjectDetectionMetric, InstanceSegmentationMetric, TextRecognitionMetric
]:
    preds = convert_to_torchmetric_format(preds, task, prediction=True)
    labels = convert_to_torchmetric_format(labels, task)

    if task == TaskType.CLASSIFICATION:
        return evaluate_classification(preds, labels, num_classes, *args, **kwargs)
    elif task == TaskType.OBJECT_DETECTION:
        return evaluate_object_detection(preds, labels, num_classes, *args, **kwargs)
    elif task == TaskType.INSTANCE_SEGMENTATION:
        return evaluate_segmentation(preds, labels, num_classes, *args, **kwargs)
    elif task == TaskType.TEXT_RECOGNITION:
        return evalute_text_recognition(preds, labels, num_classes, *args, **kwargs)
    else:
        raise NotImplementedError



import json
from dataclasses import asdict, dataclass
from pathlib import Path

import yaml
from waffle_utils.file import io


@dataclass
class BaseSchema:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        return

    def to_dict(self):
        return asdict(self)

    def save_json(self, save_path):
        d = self.to_dict()
        for k, v in d.items():
            try:
                json.dumps(v)
            except:
                d[k] = str(v)

        io.save_json(d, save_path, create_directory=True)

    def save_yaml(self, save_path):
        d = self.to_dict()
        for k, v in d.items():
            try:
                json.dumps(v)  # yaml does not catch any error, so use json instead
            except:
                d[k] = str(v)
        io.save_yaml(d, save_path, create_directory=True)

    @classmethod
    def load(cls, load_path):
        load_path = Path(load_path)
        if load_path.suffix == ".json":
            config = io.load_json(load_path)
        elif load_path.suffix == ".yaml":
            config = io.load_yaml(load_path)
        return cls(**config)



from waffle_hub import TaskType
from waffle_hub.schema.configs import TrainConfig

# Common
MODEL_TYPES = {
    TaskType.OBJECT_DETECTION: {
        "DETA": {
            "base": "jozhang97/deta-resnet-50",
        },
        "DETR": {
            "base": "facebook/detr-resnet-50",
            "large": "facebook/detr-resnet-101",
        },
        "YOLOS": {
            "tiny": "hustvl/yolos-tiny",
        },
    },
    TaskType.CLASSIFICATION: {
        "ViT": {
            "tiny": "WinKawaks/vit-tiny-patch16-224",
            "base": "google/vit-base-patch16-224",
        }
    },
}

DEFAULT_PARAMS = {
    TaskType.OBJECT_DETECTION: {
        "DETA": {
            "base": TrainConfig(
                epochs=50,
                image_size=[800, 800],
                learning_rate=5e-05,  # TODO: implement letter_box
                letter_box=True,
                batch_size=1,
            )
        },
        "DETR": {
            "base": TrainConfig(
                epochs=50,
                image_size=[800, 800],
                learning_rate=5e-05,
                letter_box=True,  # TODO: implement letter_box
                batch_size=1,
            ),
            "large": TrainConfig(
                epochs=50,
                image_size=[800, 800],
                learning_rate=5e-05,
                letter_box=True,  # TODO: implement letter_box
                batch_size=1,
            ),
        },
        "YOLOS": {
            "tiny": TrainConfig(
                epochs=50,
                image_size=[800, 800],
                learning_rate=5e-05,
                letter_box=True,  # TODO: implement letter_box
                batch_size=16,
            ),
        },
    },
    TaskType.CLASSIFICATION: {
        "ViT": {
            "tiny": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=5e-05,
                letter_box=False,
                batch_size=128,
            ),
            "base": TrainConfig(
                epochs=50,
                image_size=[224, 224],
                learning_rate=5e-05,
                letter_box=False,
                batch_size=128,
            ),
        },
    },
}



__version__ = "0.2.5"

import enum
from collections import OrderedDict

BACKEND_MAP = OrderedDict(
    {
        "ultralytics": {
            "import_path": "waffle_hub.hub.adapter.ultralytics",
            "class_name": "UltralyticsHub",
        },
        "autocare_dlt": {
            "import_path": "waffle_hub.hub.adapter.autocare_dlt",
            "class_name": "AutocareDLTHub",
        },
        "transformers": {
            "import_path": "waffle_hub.hub.adapter.transformers",
            "class_name": "TransformersHub",
        },
    }
)


class CustomEnumMeta(enum.EnumMeta):
    def __contains__(cls, item):
        if isinstance(item, str):
            return item.upper() in cls._member_names_
        return super().__contains__(item)

    def __upper__(self):
        return self.name.upper()


class BaseEnum(enum.Enum, metaclass=CustomEnumMeta):
    """Base class for Enum

    Example:
        >>> class Color(BaseEnum):
        >>>     RED = 1
        >>>     GREEN = 2
        >>>     BLUE = 3
        >>> Color.RED == "red"
        True
        >>> Color.RED == "RED"
        True
        >>> "red" in DataType
        True
        >>> "RED" in DataType
        True
    """

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name.upper() == other.upper()
        return super().__eq__(other)

    def __hash__(self):
        return hash(self.name.upper())

    def __str__(self):
        return self.name.upper()

    def __repr__(self):
        return self.name.upper()


class DataType(BaseEnum):
    # TODO: map to same value

    YOLO = enum.auto()
    ULTRALYTICS = enum.auto()

    COCO = enum.auto()

    AUTOCARE_DLT = enum.auto()

    TRANSFORMERS = enum.auto()


class TaskType(BaseEnum):
    CLASSIFICATION = enum.auto()
    OBJECT_DETECTION = enum.auto()
    SEMANTIC_SEGMENTATION = enum.auto()
    INSTANCE_SEGMENTATION = enum.auto()
    KEYPOINT_DETECTION = enum.auto()
    TEXT_RECOGNITION = enum.auto()
    REGRESSION = enum.auto()


class SplitMethod(BaseEnum):
    RANDOM = enum.auto()
    STRATIFIED = enum.auto()


EXPORT_MAP = {
    DataType.YOLO: "YOLO",
    DataType.ULTRALYTICS: "YOLO",
    DataType.COCO: "COCO",
    DataType.AUTOCARE_DLT: "AUTOCARE_DLT",
    DataType.TRANSFORMERS: "TRANSFORMERS",
}



import subprocess
import sys
from pathlib import Path
from typing import Union


def _refine_script(script_path):
    """Make script file to be compatible with both on Windows and Linux."""
    with open(script_path) as f:
        script = f.read()
    script = script.replace("\r\n", "\n").replace("\\", "/")
    with open(script_path, "w") as f:
        f.writelines(script)


def run_python_file(file_path: Union[str, Path]):
    """Run python file as a subprocess."""
    file_path = Path(file_path)

    _refine_script(file_path)

    return subprocess.run(
        [sys.executable, str(file_path)],
        check=True,
    )



"""
How to use

python -m waffle_hub.experimental.serve \
    --name test  # hub name \
    --root_dir hubs  # root directory \
    --host 0.0.0.0  # host name \
    --port 8000  # port number \
    --device 0  # cuda device id or cpu
"""
import logging

import cv2
import fastapi
import numpy as np
import torch
import uvicorn
from waffle_utils.log import initialize_logger

initialize_logger("logs/serve.log")

from waffle_hub.hub.adapter.ultralytics import UltralyticsHub
from waffle_hub.hub.model.wrapper import get_parser
from waffle_hub.schema.data import ImageInfo

app = fastapi.FastAPI()

# recive image file with post method
@app.post("/predict")
async def predict(file: fastapi.UploadFile = fastapi.File(...)):
    image = await file.read()
    image = np.frombuffer(image, dtype=np.uint8)
    image = cv2.imdecode(image, cv2.IMREAD_COLOR)
    H, W = image.shape[:2]

    image = cv2.resize(image, dsize=image_size, interpolation=cv2.INTER_CUBIC)
    image = (
        torch.from_numpy(image)
        .permute(2, 0, 1)
        .float()
        .div(255.0)
        .unsqueeze(0)
    )

    logging.info(f"start inference")

    try:
        result = model(image.to(device))
        result = parser(
            result,
            [
                ImageInfo(
                    ori_shape=(W, H),
                    new_shape=image_size,
                    input_shape=image_size,
                    pad=(0, 0),
                )
            ],
        )[0]
        print(result)
    except Exception as e:
        logging.error(f"Error: {e}")
        raise fastapi.HTTPException(status_code=500, detail=f"Error: {e}")

    logging.info(f"end inference")

    return [res.to_dict() for res in result]


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--name", type=str, required=True, help="hub name")
    parser.add_argument(
        "--root_dir", type=str, default=None, help="root directory"
    )
    parser.add_argument(
        "--host", type=str, default="0.0.0.0", help="host name"
    )
    parser.add_argument("--port", type=int, default=8000, help="port number")
    parser.add_argument(
        "--device", default="0", type=str, help="cuda device id or cpu"
    )
    args = parser.parse_args()

    device = torch.device(
        "cuda:" + args.device
        if torch.cuda.is_available() and args.device != "cpu"
        else "cpu"
    )
    logging.info(f"using device: {device}")

    hub: UltralyticsHub = UltralyticsHub.load(
        args.name, root_dir=args.root_dir
    )
    logging.info(f"loaded hub: {hub.name}")

    train_config = hub.get_train_config()
    image_size = train_config.image_size
    image_size = (
        (image_size, image_size) if isinstance(image_size, int) else image_size
    )
    parser = get_parser(hub.task)(**train_config.to_dict())

    model = hub.get_model()
    model.eval()
    model.to(device)

    uvicorn.run(app, host=args.host, port=args.port)



"""
Ultralytics Hub
See Hub documentation for more details about usage.
"""

import warnings
from pathlib import Path
from typing import Union

import torch
import ultralytics
from torchvision import transforms as T
from ultralytics import YOLO
from waffle_utils.file import io

from waffle_hub import TaskType
from waffle_hub.hub import Hub
from waffle_hub.hub.model.wrapper import ModelWrapper
from waffle_hub.schema.configs import TrainConfig
from waffle_hub.utils.callback import TrainCallback
from waffle_hub.utils.process import run_python_file

from .config import DEFAULT_PARAMS, MODEL_TYPES, TASK_MAP, TASK_SUFFIX


class UltralyticsHub(Hub):
    BACKEND_NAME = "ultralytics"
    MODEL_TYPES = MODEL_TYPES
    MULTI_GPU_TRAIN = True
    DEFAULT_PARAMS = DEFAULT_PARAMS

    TASK_MAP = TASK_MAP
    TASK_SUFFIX = TASK_SUFFIX

    def __init__(
        self,
        name: str,
        task: str = None,
        model_type: str = None,
        model_size: str = None,
        categories: Union[list[dict], list] = None,
        root_dir: str = None,
        backend: str = None,
        version: str = None,
        *args,
        **kwargs,
    ):
        if backend is not None and UltralyticsHub.BACKEND_NAME != backend:
            raise ValueError(
                f"Backend {backend} is not supported. Please use {UltralyticsHub.BACKEND_NAME}"
            )

        if version is not None and ultralytics.__version__ != version:
            warnings.warn(
                f"You've loaded the Hub created with ultralytics=={version}, \n"
                + f"but the installed version is {ultralytics.__version__}."
            )

        super().__init__(
            name=name,
            backend=UltralyticsHub.BACKEND_NAME,
            version=ultralytics.__version__,
            task=task,
            model_type=model_type,
            model_size=model_size,
            categories=categories,
            root_dir=root_dir,
        )

        self.backend_task_name = self.TASK_MAP[self.task]

    @classmethod
    def new(
        cls,
        name: str,
        task: str = None,
        model_type: str = None,
        model_size: str = None,
        categories: Union[list[dict], list] = None,
        root_dir: str = None,
        *args,
        **kwargs,
    ):
        """Create Ultralytics Hub.

        Args:
            name (str): Hub name
            task (str, optional): Task Name. See UltralyticsHub.TASKS. Defaults to None.
            model_type (str, optional): Model Type. See UltralyticsHub.MODEL_TYPES. Defaults to None.
            model_size (str, optional): Model Size. See UltralyticsHub.MODEL_SIZES. Defaults to None.
            categories (Union[list[dict], list]): class dictionary or list. [{"supercategory": "name"}, ] or ["name",].
            root_dir (str, optional): Root directory of hub repository. Defaults to None.
        """

        warnings.warn("UltralyticsHub.new() is deprecated. Please use Hub.new() instead.")

        return cls(
            name=name,
            task=task,
            model_type=model_type,
            model_size=model_size,
            categories=categories,
            root_dir=root_dir,
        )

    # Hub Utils
    def get_preprocess(self, *args, **kwargs):

        if self.task == TaskType.CLASSIFICATION:
            normalize = T.Normalize([0, 0, 0], [1, 1, 1], inplace=True)

            def preprocess(x, *args, **kwargs):
                return normalize(x)

        elif self.task == TaskType.OBJECT_DETECTION:
            normalize = T.Normalize([0, 0, 0], [1, 1, 1], inplace=True)

            def preprocess(x, *args, **kwargs):
                return normalize(x)

        elif self.task == TaskType.INSTANCE_SEGMENTATION:
            normalize = T.Normalize([0, 0, 0], [1, 1, 1], inplace=True)

            def preprocess(x, *args, **kwargs):
                return normalize(x)

        else:
            raise NotImplementedError(f"Task {self.task} is not implemented.")

        return preprocess

    def get_postprocess(self, *args, **kwargs):

        id_mapper: list[int] = kwargs.get("id_mapper", [i for i in range(len(self.categories))])

        if self.task == TaskType.CLASSIFICATION:

            def inner(x: torch.Tensor, *args, **kwargs):
                return [x[:, id_mapper]]

        elif self.task == TaskType.OBJECT_DETECTION:

            def inner(x: torch.Tensor, image_size: tuple[int, int], *args, **kwargs):
                x = x[0]  # x[0]: prediction, x[1]: TODO: what is this...?
                x = x.transpose(1, 2)

                cxcywh = x[:, :, :4]
                cx, cy, w, h = torch.unbind(cxcywh, dim=-1)
                x1 = cx - w / 2
                y1 = cy - h / 2
                x2 = cx + w / 2
                y2 = cy + h / 2
                xyxy = torch.stack([x1, y1, x2, y2], dim=-1)

                xyxy[:, :, ::2] /= image_size[0]
                xyxy[:, :, 1::2] /= image_size[1]
                probs = x[:, :, 4:][:, :, id_mapper]
                confidences, class_ids = torch.max(probs, dim=-1)

                return xyxy, confidences, class_ids

        elif self.task == TaskType.INSTANCE_SEGMENTATION:
            num_category = len(self.categories)

            def inner(x: torch.Tensor, image_size: tuple[int, int], *args, **kwargs):
                preds = x[0]  # x[0]: prediction, x[1]: TODO: what is this...?
                preds = preds.transpose(1, 2)

                probs = preds[:, :, 4 : 4 + num_category]
                confidences, class_ids = torch.max(probs, dim=-1)
                _, indicies = torch.topk(
                    confidences, k=min(100, confidences.shape[-1]), dim=-1, largest=True
                )  # TODO: make k configurable

                preds = torch.gather(preds, 1, indicies.unsqueeze(-1).repeat(1, 1, preds.shape[-1]))
                confidences = torch.gather(confidences, 1, indicies)
                class_ids = torch.gather(class_ids, 1, indicies)

                cxcywh = preds[:, :, :4]
                cx, cy, w, h = torch.unbind(cxcywh, dim=-1)
                x1 = cx - w / 2
                y1 = cy - h / 2
                x2 = cx + w / 2
                y2 = cy + h / 2
                xyxy = torch.stack([x1, y1, x2, y2], dim=-1)

                xyxy[:, :, ::2] /= image_size[0]
                xyxy[:, :, 1::2] /= image_size[1]

                # [batch, mask_dim, mask_height, mask_width] -> [batch, num_mask, mask_height * mask_width]
                protos = x[1][-1] if len(x[1]) == 3 else x[1]
                mask_size = (protos.shape[-2], protos.shape[-1])
                protos = protos.view(
                    protos.shape[0], -1, protos.shape[-2] * protos.shape[-1]
                ).float()

                masks = preds[:, :, 4 + num_category :]
                masks = torch.bmm(masks, protos).sigmoid().view(masks.shape[0], -1, *mask_size)

                return xyxy, confidences, class_ids, masks

        else:
            raise NotImplementedError(f"Task {self.task} is not implemented.")

        return inner

    def get_metrics(self) -> list[list[dict]]:
        # read csv file
        # epoch,         train/box_loss,         train/cls_loss,         train/dfl_loss,   metrics/precision(B),      metrics/recall(B),       metrics/mAP50(B),    metrics/mAP50-95(B),           val/box_loss,           val/cls_loss,           val/dfl_loss,                 lr/pg0,                 lr/pg1,                 lr/pg2
        #              0,                 2.0349,                 4.4739,                  1.214,                0.75962,                0.34442,                0.22858,                0.18438,                0.61141,                 2.8409,                0.78766,                 0.0919,                 0.0009,                 0.0009
        #              1,                 1.7328,                 4.0078,                 1.1147,                0.12267,                0.40909,                0.20891,                0.15007,                0.75082,                 2.7157,                0.79723,               0.082862,              0.0018624,              0.0018624
        # and parse it to list[dict]
        # [[{"tag": train/box_loss, "value": 0.0}, {"tag": train/box_loss, "value": 1.7328}, ...], ...]
        # and return it

        csv_path = self.artifact_dir / "results.csv"

        if not csv_path.exists():
            return []

        with open(csv_path) as f:
            lines = f.readlines()

        header = lines[0].strip().split(",")
        metrics = []
        for line in lines[1:]:
            values = line.strip().split(",")
            metric = []
            for i, value in enumerate(values):
                metric.append(
                    {
                        "tag": header[i].strip(),
                        "value": float(value),
                    }
                )
            metrics.append(metric)

        return metrics

    # Train Hook
    def on_train_start(self, cfg: TrainConfig):

        # set data
        cfg.dataset_path: Path = Path(cfg.dataset_path)
        if self.backend_task_name in ["detect", "segment"]:
            if cfg.dataset_path.suffix not in [".yml", ".yaml"]:
                yaml_files = list(cfg.dataset_path.glob("*.yaml")) + list(
                    cfg.dataset_path.glob("*.yml")
                )
                if len(yaml_files) != 1:
                    raise FileNotFoundError(f"Ambiguous data file. Detected files: {yaml_files}")
                cfg.dataset_path = Path(yaml_files[0]).absolute()
            else:
                cfg.dataset_path = cfg.dataset_path.absolute()
        elif self.backend_task_name == "classify":
            if not cfg.dataset_path.is_dir():
                raise ValueError(
                    f"Classification dataset should be directory. Not {cfg.dataset_path}"
                )
            cfg.dataset_path = cfg.dataset_path.absolute()
        cfg.dataset_path = str(cfg.dataset_path)

        # pretrained model
        cfg.pretrained_model = (
            cfg.pretrained_model
            if cfg.pretrained_model
            else self.model_type + self.model_size + self.TASK_SUFFIX[self.backend_task_name] + ".pt"
        )

    def training(self, cfg: TrainConfig, callback: TrainCallback):

        code = f"""if __name__ == "__main__":
        from ultralytics import YOLO
        model = YOLO("{cfg.pretrained_model}", task="{self.backend_task_name}")
        model.train(
            data="{cfg.dataset_path}",
            epochs={cfg.epochs},
            batch={cfg.batch_size},
            imgsz={cfg.image_size},
            lr0={cfg.learning_rate},
            lrf={cfg.learning_rate},
            rect={cfg.letter_box},
            device="{cfg.device}",
            workers={cfg.workers},
            seed={cfg.seed},
            verbose={cfg.verbose},
            project="{self.hub_dir}",
            name="{self.ARTIFACT_DIR}",
        )"""

        script_file = str((self.hub_dir / "train.py").absolute())
        with open(script_file, "w") as f:
            f.write(code)

        run_python_file(script_file)

    def on_train_end(self, cfg: TrainConfig):
        io.copy_file(
            self.artifact_dir / "weights" / "best.pt",
            self.best_ckpt_file,
            create_directory=True,
        )
        io.copy_file(
            self.artifact_dir / "weights" / "last.pt",
            self.last_ckpt_file,
            create_directory=True,
        )
        io.save_json(self.get_metrics(), self.metric_file)

    # Inference Hook
    def get_model(self):
        """Get model.
        Returns:
            ModelWrapper: Model wrapper
        """
        self.check_train_sanity()

        # get model
        model = YOLO(self.best_ckpt_file).model.eval()

        # get adapt functions
        names: list[str] = list(map(lambda x: x["name"], self.categories))
        yolo_names: dict[int, str] = model.names
        if len(names) != len(yolo_names):
            raise ValueError(
                f"Number of categories is not matched. hub: {len(names)} != ultralytics: {len(yolo_names)}"
            )

        id_mapper: list[int] = [i for i in range(len(yolo_names))]

        yolo_names_inv = {v: k for k, v in yolo_names.items()}
        for i, name in enumerate(names):
            id_mapper[i] = yolo_names_inv[name]

        preprocess = self.get_preprocess()
        postprocess = self.get_postprocess(id_mapper=id_mapper)

        # wrap model
        model = ModelWrapper(
            model=model,
            preprocess=preprocess,
            postprocess=postprocess,
        )

        return model



import argparse
from pathlib import Path

import cv2
import torch
import torchvision.transforms as T

import tqdm

from groundingdino.models import build_model
from groundingdino.util.slconfig import SLConfig
from groundingdino.util.utils import clean_state_dict, get_phrases_from_posmap

from waffle_utils.dataset import Dataset
from waffle_utils.file import io
from waffle_utils.log import initialize_logger

initialize_logger("logs/auto_label.log")

from waffle_hub.utils.data import ImageDataset
from waffle_hub.utils.draw import draw_results
from waffle_hub.schema.data import ImageInfo
from waffle_hub.schema.fields import Annotation


def parse_results(bboxes: torch.Tensor, confs: list[float], labels: list[str], image_info: ImageInfo, class2id: dict):

    W, H = image_info.input_shape
    left_pad, top_pad = image_info.pad
    ori_w, ori_h = image_info.ori_shape
    new_w, new_h = image_info.new_shape

    # cx cy w h (0~1) -> x1 y1 x2 y2 (0~1)
    cx, cy, w, h = torch.unbind(bboxes, dim=-1)
    x1 = cx - w / 2
    y1 = cy - h / 2
    x2 = cx + w / 2
    y2 = cy + h / 2
    bboxes = torch.stack([x1, y1, x2, y2], dim=-1)

    parsed = []
    for (x1, y1, x2, y2), conf, label in zip(
        bboxes, confs, labels
    ):

        x1 = max(float((x1 * W - left_pad) / new_w * ori_w), 0)
        y1 = max(float((y1 * H - top_pad) / new_h * ori_h), 0)
        x2 = min(float((x2 * W - left_pad) / new_w * ori_w), ori_w)
        y2 = min(float((y2 * H - top_pad) / new_h * ori_h), ori_h)

        parsed.append(
            Annotation.object_detection(
                bbox=[x1, y1, x2 - x1, y2 - y1],
                area=float((x2 - x1) * (y2 - y1)),
                category_id=class2id[label],
                score=float(conf),
            )
        )
    return parsed


def load_model(model_config_path, model_checkpoint_path, device):
    args = SLConfig.fromfile(model_config_path)
    args.device = device
    model = build_model(args)
    checkpoint = torch.load(model_checkpoint_path, map_location="cpu")
    load_res = model.load_state_dict(clean_state_dict(checkpoint["model"]), strict=False)
    print(load_res)
    _ = model.eval()
    return model

def get_grounding_output(model, image, caption, box_threshold, text_threshold, device, with_logits=True):
    caption = caption.lower()
    caption = caption.strip()
    if not caption.endswith("."):
        caption = caption + "."
    model = model.to(device)
    image = image.to(device)
    with torch.no_grad():
        outputs = model(image, captions=[caption])
    logits = outputs["pred_logits"].cpu().sigmoid()[0]  # (nq, 256)
    boxes = outputs["pred_boxes"].cpu()[0]  # (nq, 4)
    logits.shape[0]

    # filter output
    logits_filt = logits.clone()
    boxes_filt = boxes.clone()
    filt_mask = logits_filt.max(dim=1)[0] > box_threshold
    logits_filt = logits_filt[filt_mask]  # num_filt, 256
    boxes_filt = boxes_filt[filt_mask]  # num_filt, 4
    logits_filt.shape[0]

    # get phrase
    tokenlizer = model.tokenizer
    tokenized = tokenlizer(caption)
    # build pred
    boxes = []
    labels = []
    confs = []
    for logit, box in zip(logits_filt, boxes_filt):
        pred_phrase = get_phrases_from_posmap(logit > text_threshold, tokenized, tokenlizer)
        boxes.append(box.tolist())
        labels.append(pred_phrase)
        confs.append(logit.max().item())

    return torch.tensor(boxes), confs, labels


if __name__ == "__main__":

    parser = argparse.ArgumentParser("Grounding DINO example", add_help=True)
    parser.add_argument("--config_file", "-c", type=str, default="autolabel_tmp/GroundingDINO_SwinT_OGC.py", help="path to config file")
    parser.add_argument(
        "--checkpoint_path", "-p", type=str, default="autolabel_tmp/groundingdino_swint_ogc.pth", help="path to checkpoint file"
    )
    parser.add_argument("--source", "-s", type=str, required=True, help="path to image file or directory")
    parser.add_argument("--image_size", "-i", type=int, default=640, help="image size")
    parser.add_argument("--text_prompt", "-t", type=str, required=True, help="text prompt")
    parser.add_argument(
        "--output_dir", "-o", type=str, default="outputs", help="output directory"
    )

    parser.add_argument("--box_threshold", type=float, default=0.3, help="box threshold")
    parser.add_argument("--text_threshold", type=float, default=0.25, help="text threshold")

    parser.add_argument("--draw", action="store_true", help="draw bounding boxes")

    parser.add_argument("--waffle_dataset_name", default=None, type=str, help="waffle dataset name")

    parser.add_argument(
        "--device", default="0", type=str, help="cuda device id or cpu"
    )
    args = parser.parse_args()

    # cfg
    config_file = args.config_file  # change the path of the model config file
    checkpoint_path = args.checkpoint_path  # change the path of the model

    # text_prompt = args.text_prompt
    text_prompt = args.text_prompt
    classes = [text_prompt]
    class2id = {name: i for i, name in enumerate(classes)}
    id2class = {i: name for i, name in enumerate(classes)}
    box_threshold = args.box_threshold
    text_threshold = args.text_threshold

    # device
    device = "cpu" if args.device == "cpu" else f"cuda:{args.device}"

    # directory
    source_dir = args.source
    output_dir = Path(args.output_dir)
    draw_dir = output_dir / "draw"

    # load model
    model = load_model(config_file, checkpoint_path, device)

    # get dataloader
    dl = ImageDataset(image_dir=source_dir, image_size=args.image_size, letter_box=True).get_dataloader(1, 1)
    preprocess = T.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])

    # coco format
    coco = {
        "categories": [{"id": i+1, "name": c, "supercategory": "object"} for i, c in id2class.items()],
        "images": [],
        "annotations": [],
    }
    image_id = 1
    annotation_id = 1

    for images, image_infos in tqdm.tqdm(dl, total=len(dl)):

        images = preprocess(images)
        
        # run model
        boxes, confs, labels = get_grounding_output(
            model, images, text_prompt, box_threshold, text_threshold, device
        )
        image_info = image_infos[0]  # TODO: batch
        if len(boxes) > 0:
            results = parse_results(boxes, confs, labels, image_info, class2id)
        else:
            results = []
        
        # save result as coco format
        file_name = Path(image_info.image_path).relative_to(source_dir)
        coco["images"].append(
            {
                "id": image_id,
                "file_name": str(file_name),
                "width": image_info.ori_shape[0],
                "height": image_info.ori_shape[1],
            }
        )
        for result in results:
            coco["annotations"].append(
                {
                    "id": annotation_id,
                    "image_id": image_id,
                    "category_id": result.category_id+1,
                    "bbox": result.bbox,
                    "score": result.score,
                }
            )
            annotation_id += 1
        image_id += 1

        if args.draw:
            draw = draw_results(
                image_info.image_path,
                results,
                names=classes,
            )
            draw_path = draw_dir / file_name.with_suffix(".png")
            io.make_directory(draw_path.parent)
            cv2.imwrite(str(draw_path), draw)

    # save coco format
    io.save_json(coco, output_dir / "coco.json", create_directory=True)

    # make waffle dataset
    if args.waffle_dataset_name:
        Dataset.from_coco(
            name=args.waffle_dataset_name,
            coco_file=output_dir / "coco.json",
            coco_root_dir=source_dir,
            root_dir=output_dir,
        )
        print(f"Your waffle dataset has been saved to {output_dir}/{args.waffle_dataset_name}")



from pathlib import Path
from typing import Union

import PIL.Image
from datasets.features.image import Image as ImageFeature

from datasets import (
    ClassLabel,
    Dataset,
    DatasetDict,
    Features,
    Sequence,
    Value,
)
from waffle_hub import TaskType
from waffle_hub.schema.fields import Image


def _export_transformers_classification(
    self,
    export_dir: Path,
    train_ids: list,
    val_ids: list,
    test_ids: list,
    unlabeled_ids: list,
):
    """Export dataset to Transformers format for classification task

    Args:
        export_dir (Path): Path to export directory
        train_ids (list): List of train ids
        val_ids (list): List of validation ids
        test_ids (list): List of test ids
        unlabeled_ids (list): List of unlabeled ids
    """
    category_names = self.get_category_names()
    features = Features(
        {
            "image": ImageFeature(),
            "label": ClassLabel(names=category_names),
        }
    )

    def _export(images: list[Image]):
        for image in images:
            annotation = self.get_annotations(image.image_id)[0]
            image_path = self.raw_image_dir / image.file_name
            yield {
                "image": PIL.Image.open(image_path).convert("RGB"),
                "label": category_names[annotation.category_id - 1],
            }

    dataset = {}
    for split, image_ids in zip(
        ["train", "val", "test", "unlabeled"],
        [train_ids, val_ids, test_ids, unlabeled_ids],
    ):
        if len(image_ids) == 0:
            continue

        dataset[split] = Dataset.from_generator(
            lambda: _export(self.get_images(image_ids)), features=features
        )

    dataset = DatasetDict(dataset)
    dataset.save_to_disk(export_dir)


def _export_transformers_detection(
    self,
    export_dir: Path,
    train_ids: list,
    val_ids: list,
    test_ids: list,
    unlabeled_ids: list,
):
    """Export dataset to Transformers format for detection task

    Args:
        export_dir (Path): Path to export directory
        train_ids (list): List of train ids
        val_ids (list): List of validation ids
        test_ids (list): List of test ids
        unlabeled_ids (list): List of unlabeled ids
    """
    category_names = self.get_category_names()
    features = Features(
        {
            "image": ImageFeature(),
            "image_id": Value("int32"),
            "width": Value("int32"),
            "height": Value("int32"),
            "objects": Sequence(
                {
                    "id": Value("int32"),
                    "area": Value("int32"),
                    "category": ClassLabel(names=category_names),
                    "bbox": Sequence(Value("float32")),
                }
            ),
        }
    )

    def _export(images: list[Image]):
        for image in images:
            # objects
            annotations = self.get_annotations(image.image_id)
            objects = []
            for annotation in annotations:
                objects.append(
                    {
                        "id": annotation.annotation_id,
                        "area": annotation.area,
                        "category": category_names[annotation.category_id - 1],
                        "bbox": annotation.bbox,
                    }
                )

            # image
            image_path = self.raw_image_dir / image.file_name
            pil_image = PIL.Image.open(image_path).convert("RGB")
            yield {
                "image": pil_image,
                "image_id": image.image_id,
                "width": image.width,
                "height": image.height,
                "objects": objects,
            }

    dataset = {}
    for split, image_ids in zip(
        ["train", "val", "test", "unlabeled"],
        [train_ids, val_ids, test_ids, unlabeled_ids],
    ):
        if len(image_ids) == 0:
            continue

        dataset[split] = Dataset.from_generator(
            lambda: _export(self.get_images(image_ids)), features=features
        )

    dataset = DatasetDict(dataset)
    dataset.save_to_disk(export_dir)


def export_transformers(self, export_dir: Union[str, Path]) -> str:
    """Export dataset to Transformers format

    Args:
        export_dir (str): Path to export directory

    Returns:
        str: Path to export directory
    """
    export_dir = Path(export_dir)

    train_ids, val_ids, test_ids, _ = self.get_split_ids()

    if self.task == TaskType.CLASSIFICATION:
        _export_transformers_classification(self, export_dir, train_ids, val_ids, test_ids, [])
    elif self.task == TaskType.OBJECT_DETECTION:
        _export_transformers_detection(self, export_dir, train_ids, val_ids, test_ids, [])
    else:
        raise ValueError(f"Unsupported task type: {self.task}")

    return str(export_dir)



def get_data_config(
    data_type: str,
    image_size: list[int],
    batch_size: int,
    workers: int,
    train_coco: str,
    train_root: str,
    val_coco: str,
    val_root: str,
    test_coco: str,
    test_root: str,
):
    config = {
        "data": {
            "workers_per_gpu": workers,
            "batch_size_per_gpu": batch_size,
            "img_size": image_size,
            "train": {
                "type": data_type,
                "data_root": train_root,
                "ann": train_coco,
                "augmentation": {
                    "ColorJitter": {
                        "brightness": 0.125,
                        "contrast": 0.5,
                        "saturation": 0.5,
                        "hue": 0.1,
                    },
                    "Affine": {
                        "scale": [0.5, 1.5],
                        "translate_percent": [-0.1, 0.1],
                        "always_apply": True,
                    },
                    "HorizontalFlip": {"p": 0.5},
                    "Cutout": {
                        "p": 0.5,
                        "num_holes": 8,
                        "max_h_size": 32,
                        "max_w_size": 32,
                        "fill_value": 0,
                    },
                    "ImageNormalization": {"type": "base"},
                },
            },
            "val": {
                "type": data_type,
                "data_root": val_root,
                "ann": val_coco,
                "augmentation": {"ImageNormalization": {"type": "base"}},
            },
            "test": {
                "type": data_type,
                "data_root": test_root,
                "ann": test_coco,
                "augmentation": {"ImageNormalization": {"type": "base"}},
            },
            "cache": False,
            "single_cls": False,
        }
    }

    return config
